@@ -5,17 +5,7 @@
 module Rubinius
   class Channel
     def inspect
-      "#<Rubinius::Channel>"
-    end
-
-    def as_lock(val=nil)
-      receive
-
-      begin
-        yield
-      ensure
-        self << val
-      end
+      "#<Rubinius::Channel:0x#{object_id.to_s(16)}>"
     end
   end
 end
