@@ -6,6 +6,11 @@ module ImmediateValue
     []
   end
 
+  def public_singleton_methods
+    []
+  end
+  private :public_singleton_methods
+
   def private_singleton_methods
     []
   end
@@ -16,25 +21,13 @@ module ImmediateValue
   end
   private :protected_singleton_methods
 
-  def taint
-    self
-  end
-
-  def tainted?
-    false
-  end
-
-  def freeze
-    self
-  end
-
-  def frozen?
-    false
-  end
-
   def dup
     raise TypeError, "can't dup #{self.class.name}"
   end
+
+  def clone
+    raise TypeError, "can't clone #{self.class.name}"
+  end
 end
 
 class NilClass
