@@ -1,40 +1,82 @@
 module Kernel
-  def __method__
-    scope = Rubinius::VariableScope.of_sender
+  def Array(obj)
+    ary = Rubinius::Type.check_convert_type obj, Array, :to_ary
 
-    name = scope.method.name
+    return ary if ary
 
-    # If the name is still __block__, then it's in a script, so return nil
-    return nil if name == :__block__ or name == :__script__
-    return name
+    if array = Rubinius::Type.check_convert_type(obj, Array, :to_a)
+      array
+    else
+      [obj]
+    end
   end
+  module_function :Array
+
+  def Complex(*args)
+    Rubinius.privately do
+      Complex.convert *args
+    end
+  end
+  module_function :Complex
 
   def Float(obj)
-    raise TypeError, "can't convert nil into Float" if obj.nil?
-
     case obj
-    when Float
-      obj
     when String
-      valid_re = /^\s*[+-]?((\d+_?)*\d+(\.(\d+_?)*\d+)?|\.(\d+_?)*\d+)(\s*|([eE][+-]?(\d+_?)*\d+)\s*)$/
-
-      m = valid_re.match(obj)
-
-      if !m or !m.pre_match.empty? or !m.post_match.empty?
-        raise ArgumentError, "invalid value for Float(): #{obj.inspect}"
-      end
-      obj.convert_float
+      Rubinius::Type.coerce_string_to_float obj, true
     else
-      coerced_value = Rubinius::Type.coerce_to(obj, Float, :to_f)
-      if coerced_value.nan?
-        raise ArgumentError, "invalid value for Float(): #{coerced_value.inspect}"
-      end
-      coerced_value
+      Rubinius::Type.coerce_object_to_float obj
     end
   end
   module_function :Float
 
-  def Integer(obj)
+  ##
+  # MRI uses a macro named NUM2DBL which has essentially the same semantics as
+  # Float(), with the difference that it raises a TypeError and not a
+  # ArgumentError. It is only used in a few places (in MRI and Rubinius).
+  #--
+  # If we can, we should probably get rid of this.
+
+  def FloatValue(obj)
+    exception = TypeError.new 'no implicit conversion to float'
+
+    case obj
+    when String
+      raise exception
+    else
+      begin
+        Rubinius::Type.coerce_object_to_float obj
+      rescue
+        raise exception
+      end
+    end
+  end
+  private :FloatValue
+
+  def Hash(obj)
+    return {} if obj.nil? || obj == []
+
+    if hash = Rubinius::Type.check_convert_type(obj, Hash, :to_hash)
+      return hash
+    end
+
+    raise TypeError, "can't convert #{obj.class} into Hash"
+  end
+  module_function :Hash
+
+  def Integer(obj, base=nil)
+    if obj.kind_of? String
+      if obj.empty?
+        raise ArgumentError, "invalid value for Integer: (empty string)"
+      else
+        base ||= 0
+        return obj.to_inum(base, true)
+      end
+    end
+
+    if base
+      raise ArgumentError, "base is only valid for String values"
+    end
+
     case obj
     when Integer
       obj
@@ -44,12 +86,8 @@ module Kernel
       else
         obj.to_int
       end
-    when String
-      if obj.empty?
-        raise ArgumentError, "invalid value for Integer: (empty string)"
-      else
-        obj.to_inum(0, true)
-      end
+    when NilClass
+      raise TypeError, "can't convert nil into Integer"
     else
       # Can't use coerce_to or try_convert because I think there is an
       # MRI bug here where it will return the value without checking
@@ -65,27 +103,26 @@ module Kernel
   end
   module_function :Integer
 
-  def Array(obj)
-    ary = Rubinius::Type.try_convert obj, Array, :to_ary
-
-    return ary if ary
-
-    if obj.respond_to? :to_a
-      Rubinius::Type.coerce_to(obj, Array, :to_a)
-    else
-      [obj]
+  def Rational(a, b = 1)
+    Rubinius.privately do
+      Rational.convert a, b
     end
   end
-  module_function :Array
+  module_function :Rational
 
   def String(obj)
     return obj if obj.kind_of? String
 
-    unless obj.respond_to? :to_s
-      raise TypeError, "Unable to convert to a String"
+    unless obj.respond_to?(:to_s)
+      raise TypeError, "can't convert #{obj.class} into String"
+    end
+
+    begin
+      str = obj.to_s
+    rescue NoMethodError
+      raise TypeError, "can't convert #{obj.class} into String"
     end
 
-    str = obj.to_s
     unless str.kind_of? String
       raise TypeError, "#to_s did not return a String"
     end
@@ -109,213 +146,60 @@ module Kernel
   def StringValue(obj)
     Rubinius::Type.coerce_to obj, String, :to_str
   end
-  private :StringValue
+  module_function :StringValue
 
-  ##
-  # MRI uses a macro named NUM2DBL which has essentially the same semantics as
-  # Float(), with the difference that it raises a TypeError and not a
-  # ArgumentError. It is only used in a few places (in MRI and Rubinius).
-  #--
-  # If we can, we should probably get rid of this.
+  def __method__
+    scope = Rubinius::VariableScope.of_sender
 
-  def FloatValue(obj)
-    begin
-      Float(obj)
-    rescue
-      raise TypeError, 'no implicit conversion to float'
-    end
+    name = scope.method.name
+
+    return nil if scope.method.for_module_body?
+    # If the name is still __block__, then it's in a script, so return nil
+    return nil if name == :__block__ or name == :__script__
+    name
   end
-  private :FloatValue
+  module_function :__method__
 
-  def initialize_copy(source)
-    unless instance_of?(source.class)
-      raise TypeError, "initialize_copy should take same class object"
+  alias_method :__callee__, :__method__
+  module_function :__callee__
+
+  def __dir__
+    scope = Rubinius::ConstantScope.of_sender
+    script = scope.current_script
+    basepath = script.file_path
+    fullpath = nil
+
+    return nil unless basepath
+
+    fullpath = if script.data_path
+      script.data_path
+    else
+      Rubinius.privately do
+        File.basic_realpath(basepath)
+      end
     end
-  end
-  private :initialize_copy
 
-  def warn(warning)
-    $stderr.write "#{warning}\n" unless $VERBOSE.nil?
+    File.dirname fullpath
+  end
+  module_function :__dir__
+
+  def =~(other)
     nil
   end
-  module_function :warn
 
-  def warning(message)
-    $stderr.puts message if $VERBOSE
+  def <=>(other)
+    self == other ? 0 : nil
   end
-  module_function :warning
 
-  def exit(code=0)
-    Process.exit(code)
+  def ===(other)
+    equal?(other) || self == other
   end
-  module_function :exit
-
-  def exit!(code=1)
-    Process.exit!(code)
-  end
-  module_function :exit!
 
   def abort(msg=nil)
-    msg = StringValue(msg) if msg
     Process.abort msg
   end
   module_function :abort
 
-  def printf(target, *args)
-    case target
-    when IO
-      target.printf(*args)
-    when String
-      $stdout << Rubinius::Sprinter.get(target).call(*args)
-    else
-      raise TypeError, "The first arg to printf should be an IO or a String"
-    end
-    nil
-  end
-  module_function :printf
-
-  def sprintf(str, *args)
-    Rubinius::Sprinter.get(str).call(*args)
-  end
-
-  alias_method :format, :sprintf
-  module_function :sprintf
-  module_function :format
-
-  def puts(*a)
-    $stdout.puts(*a)
-    nil
-  end
-  module_function :puts
-
-  # For each object given, prints obj.inspect followed by the
-  # system record separator to standard output (thus, separator
-  # cannot be overridden.) Prints nothing if no objects given.
-  def p(*a)
-    return nil if a.empty?
-    a.each { |obj| $stdout.puts obj.inspect }
-    $stdout.flush
-    nil
-  end
-  module_function :p
-
-  # Object oriented version of Kernel.p
-  def display(port=$>)
-    port.write self
-  end
-
-  def print(*args)
-    args.each do |obj|
-      $stdout.write obj.to_s
-    end
-    nil
-  end
-  module_function :print
-
-  def open(path, *rest, &block)
-    path = StringValue(path)
-
-    if path.kind_of? String and path.prefix? '|'
-      return IO.popen(path[1..-1], *rest, &block)
-    end
-
-    File.open(path, *rest, &block)
-  end
-  module_function :open
-
-  def srand(seed=undefined)
-    if seed.equal? undefined
-      seed = Rubinius::Randomizer.instance.generate_seed
-    end
-
-    seed = Rubinius::Type.coerce_to seed, Integer, :to_int
-    Rubinius::Randomizer.instance.swap_seed seed
-  end
-  module_function :srand
-
-  def rand(limit=0)
-    limit = Integer(limit).abs
-
-    case limit
-    when 0
-      Rubinius::Randomizer.instance.random_float
-    when Integer
-      Rubinius::Randomizer.instance.random_integer(limit - 1)
-    else
-      raise TypeError, "Integer() returned a non-integer"
-    end
-  end
-  module_function :rand
-
-  def block_given?
-    Rubinius::VariableScope.of_sender.block != nil
-  end
-  module_function :block_given?
-
-  alias_method :iterator?, :block_given?
-  module_function :iterator?
-
-  def lambda(&prc)
-    raise ArgumentError, "block required" unless prc
-    prc.lambda_style!
-    return prc
-  end
-  alias_method :proc, :lambda
-  module_function :lambda
-  module_function :proc
-
-  def caller(start=1, exclude_kernel=true)
-    # The + 1 is to skip this frame
-    Rubinius.mri_backtrace(start + 1).map do |tup|
-      cm = tup[0]
-      line = tup[1]
-      is_block = tup[2]
-      name = tup[3]
-
-      "#{cm.active_path}:#{line}:in `#{name}'"
-    end
-  end
-  module_function :caller
-
-  def global_variables
-    Rubinius.convert_to_names Rubinius::Globals.variables
-  end
-  module_function :global_variables
-
-  def loop
-    raise LocalJumpError, "no block given" unless block_given?
-
-    begin
-      while true
-        yield
-      end
-    rescue StopIteration
-    end
-  end
-  module_function :loop
-
-  #
-  # Sleeps the current thread for +duration+ seconds.
-  #
-  def sleep(duration=undefined)
-    if duration.equal? undefined
-      duration = nil
-    elsif !duration.kind_of?(Numeric)
-      raise TypeError, 'time interval must be a numeric value'
-    end
-
-    key = :__sleep_chanel__
-    unless chan = Thread.current[key]
-      chan = Rubinius::Channel.new
-      Thread.current[key] = chan
-    end
-
-    start = Process.time
-    chan.receive_timeout duration
-    Process.time - start
-  end
-  module_function :sleep
-
   def at_exit(prc=nil, &block)
     if prc
       unless prc.respond_to?(:call)
@@ -333,83 +217,63 @@ module Kernel
   end
   module_function :at_exit
 
-  def test(cmd, file1, file2=nil)
-    case cmd
-    when ?d
-      File.directory? file1
-    when ?e
-      File.exist? file1
-    when ?f
-      File.file? file1
-    else
-      false
+  def autoload(name, file)
+    Object.autoload(name, file)
+  end
+  private :autoload
+
+  def autoload?(name)
+    Object.autoload?(name)
+  end
+  private :autoload?
+
+  def block_given?
+    Rubinius::VariableScope.of_sender.block != nil
+  end
+  module_function :block_given?
+
+  alias_method :iterator?, :block_given?
+  module_function :iterator?
+
+  def caller(start=1, exclude_kernel=true)
+    # The + 1 is to skip this frame
+    Rubinius.mri_backtrace(start + 1).map do |tup|
+      code = tup[0]
+      line = tup[1]
+      is_block = tup[2]
+      name = tup[3]
+
+      "#{code.active_path}:#{line}:in `#{name}'"
     end
   end
-  module_function :test
+  module_function :caller
 
-  def trap(sig, prc=nil, &block)
-    Signal.trap(sig, prc, &block)
-  end
-  module_function :trap
-
-  def tap
-    yield self
-    self
+  def define_singleton_method(*args, &block)
+    singleton_class.send(:define_method, *args, &block)
   end
 
-  alias_method :object_id, :__id__
-
-  def id
-    Kernel.warn "Object#id IS deprecated; use Object#object_id OR ELSE."
-    __id__
+  def display(port=$>)
+    port.write self
   end
 
-  def type
-    Kernel.warn "Object#type IS fully deprecated; use Object#class OR ELSE."
-    self.class
+  def exec(*args)
+    Process.exec(*args)
   end
+  module_function :exec
 
-  alias_method :==, :equal?
-
-  # The "sorta" operator, also known as the case equality operator.
-  # Generally while #eql? and #== are stricter, #=== is often used
-  # to denote an acceptable match or inclusion. It returns true if
-  # the match is considered to be valid and false otherwise. It has
-  # one special purpose: it is the operator used by the case expression.
-  # So in this expression:
-  #
-  #   case obj
-  #   when /Foo/
-  #     ...
-  #   when "Hi"
-  #     ...
-  #   end
-  #
-  # What really happens is that `/Foo/ === obj` is attempted and so
-  # on down until a match is found or the expression ends. The use
-  # by Regexp is very illustrative: while obj may satisfy the pattern,
-  # it may not be the only option.
-  #
-  # The default #=== operator checks if the other object is #equal?
-  # to this one (i.e., is the same object) or if #== returns true.
-  # If neither is true, false is returned instead. Many classes opt
-  # to override this behaviour to take advantage of its use in a
-  # case expression and to implement more relaxed matching semantics.
-  # Notably, the above Regexp as well as String, Module and many others.
-  def ===(other)
-    equal?(other) || self == other
+  def exit(code=0)
+    Process.exit(code)
   end
+  module_function :exit
 
-  ##
-  # Regexp matching fails by default but may be overridden by subclasses,
-  # notably Regexp and String.
-
-  def =~(other)
-    false
+  def exit!(code=1)
+    Process.exit!(code)
   end
+  module_function :exit!
 
   def extend(*modules)
-    Ruby.check_frozen
+    raise ArgumentError, "wrong number of arguments (0 for 1+)" if modules.empty?
+    Rubinius.check_frozen
 
     modules.reverse_each do |mod|
       Rubinius.privately do
@@ -425,10 +289,42 @@ module Kernel
 
   alias_method :__extend__, :extend
 
+  def getc
+    $stdin.getc
+  end
+  module_function :getc
+
+  def gets(sep=$/)
+    ARGF.gets(sep)
+  end
+  module_function :gets
+
+  def global_variables
+    Rubinius::Type.convert_to_names Rubinius::Globals.variables
+  end
+  module_function :global_variables
+
+  def initialize_dup(other)
+    initialize_copy(other)
+  end
+  private :initialize_dup
+
+  def initialize_clone(other)
+    initialize_copy(other)
+  end
+  private :initialize_clone
+
+  def initialize_copy(source)
+    unless instance_of?(Rubinius::Type.object_class(source))
+      raise TypeError, "initialize_copy should take same class object"
+    end
+  end
+  private :initialize_copy
+
   def inspect
     # The protocol here seems odd, but it's to match MRI.
 
-    ivars = instance_variables
+    ivars = __instance_variables__
 
     # If this object has no ivars, then we call to_s and return that.
     # NOTE MRI has an additional check for when to call to_s. It also does:
@@ -448,7 +344,7 @@ module Kernel
 
     Thread.recursion_guard self do
       ivars.each do |var|
-        parts << "#{var}=#{instance_variable_get(var).inspect}"
+        parts << "#{var}=#{__instance_variable_get__(var).inspect}"
       end
     end
 
@@ -458,7 +354,7 @@ module Kernel
       str = "#{prefix} #{parts.join(' ')}>"
     end
 
-    str.taint if tainted?
+    Rubinius::Type.infect(str, self)
 
     return str
   end
@@ -471,20 +367,19 @@ module Kernel
   # always false.
 
   def instance_of?(cls)
-    Ruby.primitive :object_instance_of
+    Rubinius.primitive :object_instance_of
 
-    if cls.class != Class and cls.class != Module
+    arg_class = Rubinius::Type.object_class(cls)
+    if arg_class != Class and arg_class != Module
       # We can obviously compare against Modules but result is always false
       raise TypeError, "instance_of? requires a Class argument"
     end
 
-    self.class == cls
+    Rubinius::Type.object_class(self) == cls
   end
 
-  alias_method :__instance_of__, :instance_of?
-
   def instance_variable_get(sym)
-    Ruby.primitive :object_get_ivar
+    Rubinius.primitive :object_get_ivar
 
     sym = Rubinius::Type.ivar_validate sym
     instance_variable_get sym
@@ -493,7 +388,7 @@ module Kernel
   alias_method :__instance_variable_get__, :instance_variable_get
 
   def instance_variable_set(sym, value)
-    Ruby.primitive :object_set_ivar
+    Rubinius.primitive :object_set_ivar
 
     sym = Rubinius::Type.ivar_validate sym
     instance_variable_set sym, value
@@ -501,39 +396,17 @@ module Kernel
 
   alias_method :__instance_variable_set__, :instance_variable_set
 
-  def remove_instance_variable(sym)
-    Ruby.primitive :object_del_ivar
-
-    # If it's already a symbol, then we're here because it doesn't exist.
-    if sym.kind_of? Symbol
-      raise NameError, "instance variable '#{sym}' not defined"
-    end
-
-    # Otherwise because sym isn't a symbol, coerce it and try again.
-    remove_instance_variable Rubinius::Type.ivar_validate(sym)
-  end
-  private :remove_instance_variable
-
-  def all_instance_variables
-    Ruby.primitive :object_ivar_names
-
-    raise PrimitiveFailure, "Object#instance_variables failed"
-  end
-  private :all_instance_variables
-
   def instance_variables
     ary = []
-    all_instance_variables.each do |sym|
-      ary << sym.to_s if sym.is_ivar?
+    __all_instance_variables__.each do |sym|
+      ary << sym if sym.is_ivar?
     end
 
-    return ary
+    Rubinius::Type.convert_to_names ary
   end
 
-  alias_method :__instance_variables__, :instance_variables
-
   def instance_variable_defined?(name)
-    Ruby.primitive :object_ivar_defined
+    Rubinius.primitive :object_ivar_defined
 
     instance_variable_defined? Rubinius::Type.ivar_validate(name)
   end
@@ -543,37 +416,67 @@ module Kernel
   alias_method :__instance_variable_defined_p__, :instance_variable_defined?
   alias_method :__respond_to_p__, :respond_to?
 
-  def singleton_method_added(name)
-  end
-  private :singleton_method_added
-
-  def singleton_method_removed(name)
-  end
-  private :singleton_method_removed
-
-  def singleton_method_undefined(name)
-  end
-  private :singleton_method_undefined
-
   alias_method :is_a?, :kind_of?
 
+  def lambda
+    env = nil
+
+    Rubinius.asm do
+      push_block
+      # assign a pushed block to the above local variable "env"
+      # Note that "env" is indexed at 0.
+      set_local 0
+    end
+
+    raise ArgumentError, "block required" unless env
+
+    prc = Proc.__from_block__(env)
+
+    # Make a proc lambda only when passed an actual block (ie, not using the
+    # "&block" notation), otherwise don't modify it at all.
+    prc.lambda_style! if env.is_a?(Rubinius::BlockEnvironment)
+
+    return prc
+  end
+  module_function :lambda
+
+  def load(name, wrap=false)
+    cl = Rubinius::CodeLoader.new(name)
+    cl.load(wrap)
+
+    Rubinius.run_script cl.compiled_code
+
+    Rubinius::CodeLoader.loaded_hook.trigger!(name)
+
+    return true
+  end
+  module_function :load
+
+  def loop
+    return to_enum(:loop) unless block_given?
+
+    begin
+      while true
+        yield
+      end
+    rescue StopIteration
+    end
+  end
+  module_function :loop
+
   def method(name)
     name = Rubinius::Type.coerce_to_symbol name
-    cm = Rubinius.find_method(self, name)
+    code = Rubinius.find_method(self, name)
 
-    if cm
-      return Method.new(self, cm[1], cm[0], name)
+    if code
+      Method.new(self, code[1], code[0], name)
+    elsif respond_to_missing?(name, true)
+      Method.new(self, self.class, Rubinius::MissingMethod.new(self,  name), name)
     else
       raise NameError, "undefined method `#{name}' for #{self.inspect}"
     end
   end
 
-  def nil?
-    false
-  end
-
-  alias_method :__nil__, :nil?
-
   def methods(all=true)
     methods = singleton_methods(all)
 
@@ -582,7 +485,7 @@ module Kernel
       # Type.object_singleton_class raises a TypeError.
       case self
       when Fixnum, Symbol
-        methods |= self.class.instance_methods(true)
+        methods |= Rubinius::Type.object_class(self).instance_methods(true)
       else
         methods |= Rubinius::Type.object_singleton_class(self).instance_methods(true)
       end
@@ -598,8 +501,67 @@ module Kernel
     return methods - undefs
   end
 
+  def nil?
+    false
+  end
+
+  def object_id
+    Rubinius.primitive :object_id
+    raise PrimitiveFailure, "Kernel#object_id primitive failed"
+  end
+
+  def open(obj, *rest, &block)
+    if obj.respond_to?(:to_open)
+      obj = obj.to_open(*rest)
+
+      if block_given?
+        return yield(obj)
+      else
+        return obj
+      end
+    end
+
+    path = Rubinius::Type.coerce_to_path obj
+
+    if path.kind_of? String and path.prefix? '|'
+      return IO.popen(path[1..-1], *rest, &block)
+    end
+
+    File.open(path, *rest, &block)
+  end
+  module_function :open
+
+  def p(*a)
+    return nil if a.empty?
+    a.each { |obj| $stdout.puts obj.inspect }
+    $stdout.flush
+
+    a.size == 1 ? a.first : a
+  end
+  module_function :p
+
+  def print(*args)
+    args.each do |obj|
+      $stdout.write obj.to_s
+    end
+    nil
+  end
+  module_function :print
+
+  def printf(target, *args)
+    if target.kind_of?(String)
+      output = $stdout
+    else
+      output = target
+      target = args.shift
+    end
+    output.write Rubinius::Sprinter.get(target).call(*args)
+    nil
+  end
+  module_function :printf
+
   def private_methods(all=true)
-    private_singleton_methods() | self.class.private_instance_methods()
+    private_singleton_methods() | Rubinius::Type.object_class(self).private_instance_methods(all)
   end
 
   def private_singleton_methods
@@ -617,12 +579,18 @@ module Kernel
       methods.concat m.method_table.private_names
     end
 
-    Rubinius.convert_to_names methods
+    Rubinius::Type.convert_to_names methods
   end
   private :private_singleton_methods
 
+  def proc(&prc)
+    raise ArgumentError, "block required" unless prc
+    return prc
+  end
+  module_function :proc
+
   def protected_methods(all=true)
-    protected_singleton_methods() | self.class.protected_instance_methods(all)
+    protected_singleton_methods() | Rubinius::Type.object_class(self).protected_instance_methods(all)
   end
 
   def protected_singleton_methods
@@ -638,14 +606,222 @@ module Kernel
       methods.concat m.method_table.protected_names
     end
 
-    Rubinius.convert_to_names methods
+    Rubinius::Type.convert_to_names methods
   end
   private :protected_singleton_methods
 
-  def public_methods(all=true)
-    singleton_methods(all) | self.class.public_instance_methods(all)
+  def public_method(name)
+    name = Rubinius::Type.coerce_to_symbol name
+    code = Rubinius.find_public_method(self, name)
+
+    if code
+      Method.new(self, code[1], code[0], name)
+    elsif respond_to_missing?(name, false)
+      Method.new(self, self.class, Rubinius::MissingMethod.new(self,  name), name)
+    else
+      raise NameError, "undefined method `#{name}' for #{self.inspect}"
+    end
   end
 
+  def public_methods(all=true)
+    public_singleton_methods | Rubinius::Type.object_class(self).public_instance_methods(all)
+  end
+
+  def public_send(message, *args)
+    Rubinius.primitive :object_public_send
+    raise PrimitiveFailure, "Kernel#public_send primitive failed"
+  end
+
+  def public_singleton_methods
+    m = Rubinius::Type.object_singleton_class self
+    methods = m.method_table.public_names
+
+    while m = m.direct_superclass
+      unless Rubinius::Type.object_kind_of?(m, Rubinius::IncludedModule) or
+             Rubinius::Type.singleton_class_object(m)
+        break
+      end
+
+      methods.concat m.method_table.public_names
+    end
+
+    Rubinius::Type.convert_to_names methods
+  end
+  private :public_singleton_methods
+
+  def putc(int)
+    $stdout.putc(int)
+  end
+  module_function :putc
+
+  def puts(*a)
+    $stdout.puts(*a)
+    nil
+  end
+  module_function :puts
+
+  def rand(limit=0)
+    if limit == 0
+      return Thread.current.randomizer.random_float
+    end
+
+    if limit.kind_of?(Range)
+      return Thread.current.randomizer.random(limit)
+    else
+      limit = Integer(limit).abs
+
+      if limit == 0
+        Thread.current.randomizer.random_float
+      else
+        Thread.current.randomizer.random_integer(limit - 1)
+      end
+    end
+  end
+  module_function :rand
+
+  def readline(sep=$/)
+    ARGF.readline(sep)
+  end
+  module_function :readline
+
+  def readlines(sep=$/)
+    ARGF.readlines(sep)
+  end
+  module_function :readlines
+
+  def remove_instance_variable(sym)
+    Rubinius.primitive :object_del_ivar
+
+    # If it's already a symbol, then we're here because it doesn't exist.
+    if sym.kind_of? Symbol
+      raise NameError, "instance variable '#{sym}' not defined"
+    end
+
+    # Otherwise because sym isn't a symbol, coerce it and try again.
+    remove_instance_variable Rubinius::Type.ivar_validate(sym)
+  end
+
+  def require(name)
+    Rubinius::CodeLoader.require name
+  end
+  module_function :require
+
+  def require_relative(name)
+    scope = Rubinius::ConstantScope.of_sender
+    Rubinius::CodeLoader.require_relative(name, scope)
+  end
+  module_function :require_relative
+
+  def select(*args)
+    IO.select(*args)
+  end
+  module_function :select
+
+  def send(message, *args)
+    Rubinius.primitive :object_send
+    raise PrimitiveFailure, "Kernel#send primitive failed"
+  end
+
+  def set_trace_func(*args)
+    raise NotImplementedError
+  end
+  module_function :set_trace_func
+
+  def sprintf(str, *args)
+    Rubinius::Sprinter.get(str).call(*args)
+  end
+  module_function :sprintf
+
+  alias_method :format, :sprintf
+  module_function :format
+
+  def sleep(duration=undefined)
+    Rubinius.primitive :vm_sleep
+
+    # The primitive will fail on arg count if sleep is called
+    # without an argument, so we call it again passing undefined
+    # to mean "sleep forever"
+    #
+    if undefined.equal? duration
+      return sleep(undefined)
+    end
+
+    if duration.kind_of? Numeric
+      float = Rubinius::Type.coerce_to duration, Float, :to_f
+      return sleep(float)
+    else
+      raise TypeError, 'time interval must be a numeric value'
+    end
+  end
+  module_function :sleep
+
+  def srand(seed=undefined)
+    if undefined.equal? seed
+      seed = Thread.current.randomizer.generate_seed
+    end
+
+    seed = Rubinius::Type.coerce_to seed, Integer, :to_int
+    Thread.current.randomizer.swap_seed seed
+  end
+  module_function :srand
+
+  def tap
+    yield self
+    self
+  end
+
+  def test(cmd, file1, file2=nil)
+    case cmd
+    when ?d
+      File.directory? file1
+    when ?e
+      File.exist? file1
+    when ?f
+      File.file? file1
+    when ?l
+      File.symlink? file1
+    when ?r
+      File.readable? file1
+    when ?R
+      File.readable_real? file1
+    when ?w
+      File.writable? file1
+    when ?W
+      File.writable_real? file1
+    when ?A
+      File.atime file1
+    when ?C
+      File.ctime file1
+    when ?M
+      File.mtime file1
+    else
+      raise NotImplementedError, "command ?#{cmd.chr} not implemented"
+    end
+  end
+  module_function :test
+
+  def to_enum(method=:each, *args)
+    Enumerator.new(self, method, *args)
+  end
+  alias_method :enum_for, :to_enum
+
+  def trap(sig, prc=nil, &block)
+    Signal.trap(sig, prc, &block)
+  end
+  module_function :trap
+
+  def singleton_method_added(name)
+  end
+  private :singleton_method_added
+
+  def singleton_method_removed(name)
+  end
+  private :singleton_method_removed
+
+  def singleton_method_undefined(name)
+  end
+  private :singleton_method_undefined
+
   def singleton_methods(all=true)
     m = Rubinius::Type.object_singleton_class self
     mt = m.method_table
@@ -664,146 +840,18 @@ module Kernel
       end
     end
 
-    Rubinius.convert_to_names methods.uniq
+    Rubinius::Type.convert_to_names methods.uniq
   end
 
-  alias_method :send, :__send__
-
-  def to_a
-    if self.kind_of? Array
-      self
-    else
-      [self]
-    end
-  end
-
-  def to_s
-    str = "#<#{self.class}:0x#{self.__id__.to_s(16)}>"
-    str.taint if tainted?
-    return str
-  end
-
-  ##
-  # Loads the given file as executable code and returns true. If
-  # the file cannot be found, cannot be compiled or some other
-  # error occurs, LoadError is raised with an explanation.
-  #
-  # Unlike #require, the file extension if any must be present but
-  # is not restricted to .rb, .rbc or .<platform shared lib ext>.
-  # Any other extensions (or no extension) are assumed to be plain
-  # Ruby files. The only exceptions to this rule are:
-  #
-  # 1.  if given a .rb or no/any-extensioned file and there is a
-  #     compiled version of the same file that is not older than
-  #     the source file (based on File.mtime), the compiled one
-  #     is loaded directly to avoid the compilation overhead.
-  # 2.  if a .rb file is given but it does not exist, the system
-  #     will try to load the corresponding .rbc instead (to allow
-  #     distributing just .rbc files.)
-  #
-  # If the path given starts with ./, ../, ~/ or /, it is treated
-  # as a "qualified" file and will be loaded directly (after path
-  # expansion) instead of matching against $LOAD_PATH. The relative
-  # paths use Dir.pwd.
-  #
-  # If the filename is plain (unqualified) then it is sequentially
-  # prefixed with each path in $LOAD_PATH ($:) to locate the file,
-  # using the first one that exists. If none of the resulting paths
-  # exist, LoadError is raised. Unqualified names may contain path
-  # elements so directories are valid targets and can be used with
-  # $LOAD_PATH.
-  #
-  # A few extra options are supported. If the second parameter is
-  # true, then the module is wrapped inside an anonymous module for
-  # loading to avoid polluting the namespace. This is actually a
-  # shorthand for passing in :wrap => true-ish in the second arg
-  # which may be an option Hash.
-  #
-  # If :recompile in option Hash is true-ish then the file in
-  # question is recompiled each time. If the source file is not
-  # present when recompiling is requested, a LoadError is raised.
-  #
-  # TODO: Support non-UNIX paths.
-  #
-  def load(name, wrap=false)
-    cl = Rubinius::CodeLoader.new(name)
-    cl.load(wrap)
-
-    Rubinius.run_script cl.cm
-
-    Rubinius::CodeLoader.loaded_hook.trigger!(name)
-
-    return true
-  end
-  module_function :load
-
-  # Attempt to load the given file, returning true if successful.
-  # If the file has already been successfully loaded and exists
-  # in $LOADED_FEATURES, it will not be re-evaluated and false
-  # is returned instead. If the filename cannot be resolved,
-  # a LoadError is raised.
-  #
-  # The file can have one of the following extensions:
-  #
-  # [.rb]                   Plain Ruby source file.
-  # [.rbc]                  Compiled Ruby source file.
-  # [.o, .so, .dylib, .dll] Shared library (platform-specific.)
-  # [<none>]                Filename without extension.
-  #
-  # (.rba files should be loaded using CodeArchive.load_everything.)
-  #
-  # If the file does not have an extension, #require attempts to
-  # match it using .rb, .rbc and .<shared extension> as extensions,
-  # in that order, instead. If foo.rb does not exist but foo.rbc
-  # does, the latter will be loaded even if called with foo.rb.
-  #
-  # If the path given starts with ./, ../, ~/ or /, it is treated
-  # as a "qualified" file and will be loaded directly (after path
-  # expansion) instead of matching against $LOAD_PATH. The relative
-  # paths use Dir.pwd.
-  #
-  # If the filename is plain (unqualified) then it is sequentially
-  # prefixed with each path in $LOAD_PATH ($:) to locate the file,
-  # using the first one that exists. If none of the resulting paths
-  # exist, LoadError is raised. Unqualified names may contain path
-  # elements so directories are valid targets and can be used with
-  # $LOAD_PATH.
-  #
-  # TODO: Support non-UNIX paths.
-  #
-  # TODO: See if we can safely use 1.9 rules with $LOADED_FEATURES,
-  #       i.e. expand paths into it. This should be possible if it
-  #       is completely transparent to the user in all normal cases.
-  #
-  # Each successfully loaded file is added to $LOADED_FEATURES
-  # ($"), using the original unexpanded filename (with the
-  # exception that the file extension is added.)
-  #
-  def require(name)
-    Rubinius::CodeLoader.require name
-  end
-  module_function :require
-
-  def autoload(name, file)
-    Object.autoload(name, file)
-  end
-  private :autoload
-
-  def autoload?(name)
-    Object.autoload?(name)
-  end
-  private :autoload?
-
-  def set_trace_func(*args)
-    raise NotImplementedError
-  end
-  module_function :set_trace_func
-
   def syscall(*args)
     raise NotImplementedError
   end
   module_function :syscall
 
+  def to_s
+    Rubinius::Type.infect("#<#{self.class}:0x#{self.__id__.to_s(16)}>", self)
+  end
+
   def trace_var(*args)
     raise NotImplementedError
   end
@@ -814,96 +862,15 @@ module Kernel
   end
   module_function :untrace_var
 
-  # Perlisms.
-
-  def chomp(string=$/)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_ = $_.chomp(string)
+  def warn(*messages)
+    $stderr.puts(*messages) if !$VERBOSE.nil? && !messages.empty?
+    nil
   end
-  module_function :chomp
+  module_function :warn
 
-  def chomp!(string=$/)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_.chomp!(string)
+  def warning(message)
+    $stderr.puts message if $VERBOSE
   end
-  module_function :chomp!
-
-  def chop(string=$/)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_ = $_.chop(string)
-  end
-  module_function :chop
-
-  def chop!(string=$/)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_.chop!(string)
-  end
-  module_function :chop!
-
-  def getc
-    $stdin.getc
-  end
-  module_function :getc
-
-  def putc(int)
-    $stdout.putc(int)
-  end
-  module_function :putc
-
-  def gets(sep=$/)
-    ARGF.gets(sep)
-  end
-  module_function :gets
-
-  def readline(sep=$/)
-    ARGF.readline(sep)
-  end
-  module_function :readline
-
-  def readlines(sep=$/)
-    ARGF.readlines(sep)
-  end
-  module_function :readlines
-
-  def gsub(pattern, rep=nil, &block)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_ = $_.gsub(pattern, rep, &block)
-  end
-  module_function :gsub
-
-  def gsub!(pattern, rep=nil, &block)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_.gsub!(pattern, rep, &block)
-  end
-  module_function :gsub!
-
-  def sub(pattern, rep=nil, &block)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_ = $_.sub(pattern, rep, &block)
-  end
-  module_function :sub
-
-  def sub!(pattern, rep=nil, &block)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_.sub!(pattern, rep, &block)
-  end
-  module_function :sub!
-
-  def scan(pattern, &block)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_.scan(pattern, &block)
-  end
-  module_function :scan
-
-  def select(*args)
-    IO.select(*args)
-  end
-  module_function :select
-
-  def split(*args)
-    raise TypeError, "$_ must be a String" unless $_.kind_of? String
-    $_.split(*args)
-  end
-  module_function :split
+  module_function :warning
 end
 
