@@ -1,13 +1,187 @@
-class Object
+class BasicObject
   def __marshal__(ms, strip_ivars = false)
     out = ms.serialize_extended_object self
     out << "o"
-    name = self.__class__.__name__.to_sym
-    out << ms.serialize(name)
+    cls = Rubinius::Type.object_class self
+    name = Rubinius::Type.module_inspect cls
+    out << ms.serialize(name.to_sym)
     out << ms.serialize_instance_variables_suffix(self, true, strip_ivars)
   end
 end
 
+class Class
+  def __marshal__(ms)
+    if Rubinius::Type.singleton_class_object(self)
+      raise TypeError, "singleton class can't be dumped"
+    elsif name.nil? || name.empty?
+      raise TypeError, "can't dump anonymous module #{self}"
+    end
+
+    "c#{ms.serialize_integer(name.length)}#{name}"
+  end
+end
+
+class Module
+  def __marshal__(ms)
+    raise TypeError, "can't dump anonymous module #{self}" if name.nil? || name.empty?
+    "m#{ms.serialize_integer(name.length)}#{name}"
+  end
+end
+
+class Float
+  def __marshal__(ms)
+    if nan?
+      str = "nan"
+    elsif zero?
+      str = (1.0 / self) < 0 ? '-0' : '0'
+    elsif infinite?
+      str = self < 0 ? "-inf" : "inf"
+    else
+      s, decimal, sign, digits = dtoa
+
+      if decimal < -3 or decimal > digits
+        str = s.insert(1, ".") << "e#{decimal - 1}"
+      elsif decimal > 0
+        str = s[0, decimal]
+        digits -= decimal
+        str << ".#{s[decimal, digits]}" if digits > 0
+      else
+        str = "0."
+        str << "0" * -decimal if decimal != 0
+        str << s[0, digits]
+      end
+    end
+
+    Rubinius::Type.binary_string("f#{ms.serialize_integer(str.length)}#{str}")
+  end
+end
+
+class Exception
+  def __marshal__(ms)
+    out = ms.serialize_extended_object self
+    out << "o"
+    cls = Rubinius::Type.object_class self
+    name = Rubinius::Type.module_inspect cls
+    out << ms.serialize(name.to_sym)
+    out << ms.serialize_fixnum(2)
+
+    out << ms.serialize(:mesg)
+    out << ms.serialize(@reason_message)
+    out << ms.serialize(:bt)
+    out << ms.serialize(backtrace)
+
+    out
+  end
+end
+
+class Time
+  def __custom_marshal__(ms)
+    out = Rubinius::Type.binary_string("")
+
+    # Order matters.
+    extra_values = {}
+    extra_values[:offset] = gmt_offset unless gmt?
+    extra_values[:zone] = zone.encode(Encoding.find('locale'))
+
+    if nsec > 0
+      # MRI serializes nanoseconds as a Rational using an
+      # obscure and implementation-dependent method.
+      # To keep compatibility we can just put nanoseconds
+      # in the numerator and set the denominator to 1.
+      extra_values[:nano_num] = nsec
+      extra_values[:nano_den] = 1
+    end
+
+    ivars = ms.serializable_instance_variables(self, false)
+    out << Rubinius::Type.binary_string("I")
+    out << Rubinius::Type.binary_string("u#{ms.serialize(self.class.name.to_sym)}")
+
+    str = _dump
+    out << ms.serialize_integer(str.length) + str
+
+    count = ivars.size + extra_values.size
+    out << ms.serialize_integer(count)
+
+    ivars.each do |ivar|
+      sym = ivar.to_sym
+      val = __instance_variable_get__(sym)
+      out << ms.serialize(sym)
+      out << ms.serialize(val)
+    end
+
+    extra_values.each_pair do |key, value|
+      out << ms.serialize(key)
+      out << ms.serialize(value)
+    end
+
+    out
+  end
+end
+
+module Marshal
+  class State
+    def serialize_encoding?(obj)
+      enc = Rubinius::Type.object_encoding(obj)
+      enc && enc != Encoding::BINARY
+    end
+
+    def serialize_encoding(obj)
+      case enc = Rubinius::Type.object_encoding(obj)
+        when Encoding::US_ASCII
+          :E.__marshal__(self) + false.__marshal__(self)
+        when Encoding::UTF_8
+          :E.__marshal__(self) + true.__marshal__(self)
+        else
+          :encoding.__marshal__(self) + serialize_string(enc.name)
+      end
+    end
+
+    def set_object_encoding(obj, enc)
+      case obj
+      when String
+        obj.force_encoding enc
+      when Regexp
+        obj.source.force_encoding enc
+      when Symbol
+        # TODO
+      end
+    end
+
+    def set_instance_variables(obj)
+      construct_integer.times do
+        ivar = get_symbol
+        value = construct
+
+        case ivar
+        when :E
+          if value
+            set_object_encoding obj, Encoding::UTF_8
+          else
+            set_object_encoding obj, Encoding::US_ASCII
+          end
+          next
+        when :encoding
+          if enc = Encoding.find(value)
+            set_object_encoding obj, enc
+            next
+          end
+        end
+
+        obj.__instance_variable_set__ prepare_ivar(ivar), value
+      end
+    end
+
+    def construct_string
+      obj = get_byte_sequence
+      Rubinius::Unsafe.set_class(obj, get_user_class) if @user_class
+
+      set_object_encoding(obj, Encoding::ASCII_8BIT)
+
+      store_unique_object obj
+    end
+  end
+end
+
 class Range
   def __marshal__(ms)
     super(ms, true)
@@ -16,50 +190,29 @@ end
 
 class NilClass
   def __marshal__(ms)
-    "0"
+    Rubinius::Type.binary_string("0")
   end
 end
 
 class TrueClass
   def __marshal__(ms)
-    "T"
+    Rubinius::Type.binary_string("T")
   end
 end
 
 class FalseClass
   def __marshal__(ms)
-    "F"
-  end
-end
-
-class Class
-  def __marshal__(ms)
-    if Rubinius::Type.singleton_class_object(self)
-      raise TypeError, "singleton class can't be dumped"
-    elsif name.empty?
-      raise TypeError, "can't dump anonymous module #{self}"
-    end
-
-    "c#{ms.serialize_integer(name.length)}#{name}"
-  end
-end
-
-class Module
-  def __marshal__(ms)
-    raise TypeError, "can't dump anonymous module #{self}" if name.empty?
-    "m#{ms.serialize_integer(name.length)}#{name}"
+    Rubinius::Type.binary_string("F")
   end
 end
 
 class Symbol
   def __marshal__(ms)
-    if idx = ms.find_symlink(self) then
-      ";#{ms.serialize_integer(idx)}"
+    if idx = ms.find_symlink(self)
+      Rubinius::Type.binary_string(";#{ms.serialize_integer(idx)}")
     else
       ms.add_symlink self
-
-      str = to_s
-      ":#{ms.serialize_integer(str.length)}#{str}"
+      ms.serialize_symbol(self)
     end
   end
 end
@@ -69,10 +222,8 @@ class String
     out =  ms.serialize_instance_variables_prefix(self)
     out << ms.serialize_extended_object(self)
     out << ms.serialize_user_class(self, String)
-    out << '"'
-    out << ms.serialize_integer(self.length) << self
+    out << ms.serialize_string(self)
     out << ms.serialize_instance_variables_suffix(self)
-
     out
   end
 end
@@ -97,7 +248,7 @@ class Regexp
     out << ms.serialize_user_class(self, Regexp)
     out << "/"
     out << ms.serialize_integer(str.length) + str
-    out << (options & 0x7).chr
+    out << (options & Regexp::OPTION_MASK).chr
     out << ms.serialize_instance_variables_suffix(self)
 
     out
@@ -106,7 +257,8 @@ end
 
 class Struct
   def __marshal__(ms)
-    exclude = _attrs.map { |a| "@#{a}" }
+    attr_syms = _attrs.map { |a| "@#{a}".to_sym }
+    exclude = Rubinius::Type.convert_to_names attr_syms
 
     out =  ms.serialize_instance_variables_prefix(self, exclude)
     out << ms.serialize_extended_object(self)
@@ -134,7 +286,7 @@ class Array
     out << ms.serialize_user_class(self, Array)
     out << "["
     out << ms.serialize_integer(self.length)
-    unless empty? then
+    unless empty?
       each do |element|
         out << ms.serialize(element)
       end
@@ -149,16 +301,19 @@ class Hash
   def __marshal__(ms)
     raise TypeError, "can't dump hash with default proc" if default_proc
 
-    #excluded_ivars = %w[@bins @count @records]
-    excluded_ivars = %w[@capacity @mask @max_entries @size @entries @default_proc @default]
+    exclude_syms = %w[
+      @capacity @mask @max_entries @size @entries @default_proc @default
+      @state @compare_by_identity @head @tail @table
+    ].map { |a| a.to_sym }
+    excluded_ivars = Rubinius::Type.convert_to_names exclude_syms
 
     out =  ms.serialize_instance_variables_prefix(self, excluded_ivars)
     out << ms.serialize_extended_object(self)
     out << ms.serialize_user_class(self, Hash)
     out << (self.default ? "}" : "{")
     out << ms.serialize_integer(length)
-    unless empty? then
-      each_pair do |(key, val)|
+    unless empty?
+      each_pair do |key, val|
         out << ms.serialize(key)
         out << ms.serialize(val)
       end
@@ -171,18 +326,23 @@ class Hash
   end
 end
 
-class Float
-  def __marshal__(ms)
-    str = if nan? then
-            "nan"
-          elsif zero? then
-            (1.0 / self) < 0 ? '-0' : '0'
-          elsif infinite? then
-            self < 0 ? "-inf" : "inf"
-          else
-            ("%.*g" % [17, self]) + ms.serialize_mantissa(self)
-          end
-    "f#{ms.serialize_integer(str.length)}#{str}"
+class Time
+  def self.__construct__(ms, data, ivar_index, has_ivar)
+    obj = _load(data)
+    ms.store_unique_object obj
+
+    if ivar_index and has_ivar[ivar_index]
+      ms.set_instance_variables obj
+      has_ivar[ivar_index] = false
+    end
+
+    nano_num = obj.instance_variable_get(:@nano_num)
+    nano_den = obj.instance_variable_get(:@nano_den)
+    if nano_num && nano_den
+      obj.send(:nsec=, Rational(nano_num, nano_den).to_i)
+    end
+
+    obj
   end
 end
 
@@ -260,7 +420,6 @@ module Marshal
       # loading
       if stream
         @stream = stream
-        @byte_array = stream.data
       else
         @stream = nil
       end
@@ -277,11 +436,21 @@ module Marshal
       @call = true
     end
 
-    def const_lookup(name)
+    def const_lookup(name, type = nil)
       mod = Object
 
       parts = String(name).split '::'
-      parts.each { |part| mod = mod.const_get(part) }
+      parts.each do |part|
+        unless Rubinius::Type.const_exists?(mod, part)
+          raise ArgumentError, "undefined class/module #{name}"
+        end
+
+        mod = Rubinius::Type.const_get(mod, part, false)
+      end
+
+      if type and not mod.instance_of? type
+        raise ArgumentError, "#{name} does not refer to a #{type}"
+      end
 
       mod
     end
@@ -307,79 +476,78 @@ module Marshal
     def construct(ivar_index = nil, call_proc = true)
       type = consume_byte()
       obj = case type
-            when ?0
+            when 48   # ?0
               nil
-            when ?T
+            when 84   # ?T
               true
-            when ?F
+            when 70   # ?F
               false
-            when ?c, ?m
-              # Don't use construct_symbol, because we must not
-              # memoize this symbol.
-              name = get_byte_sequence.to_sym
-              obj = const_lookup name
-
-              store_unique_object obj
-
-              obj
-            when ?i
+            when 99   # ?c
+              construct_class
+            when 109  # ?m
+              construct_module
+            when 77   # ?M
+              construct_old_module
+            when 105  # ?i
               construct_integer
-            when ?l
+            when 108  # ?l
               construct_bignum
-            when ?f
+            when 102  # ?f
               construct_float
-            when ?:
+            when 58   # ?:
               construct_symbol
-            when ?"
+            when 34   # ?"
               construct_string
-            when ?/
+            when 47   # ?/
               construct_regexp
-            when ?[
+            when 91   # ?[
               construct_array
-            when ?{
+            when 123  # ?{
               construct_hash
-            when ?}
+            when 125  # ?}
               construct_hash_def
-            when ?S
+            when 83   # ?S
               construct_struct
-            when ?o
+            when 111  # ?o
               construct_object
-            when ?u
+            when 117  # ?u
               construct_user_defined ivar_index
-            when ?U
+            when 85   # ?U
               construct_user_marshal
-            when ?@
+            when 100  # ?d
+              construct_data
+            when 64   # ?@
               num = construct_integer
               obj = @objects[num]
 
               raise ArgumentError, "dump format error (unlinked)" unless obj
 
               return obj
-            when ?;
+            when 59   # ?;
               num = construct_integer
               sym = @symbols[num]
 
               raise ArgumentError, "bad symbol" unless sym
 
               return sym
-            when ?e
+            when 101  # ?e
               @modules ||= []
 
               name = get_symbol
-              @modules << const_lookup(name)
+              @modules << const_lookup(name, Module)
 
               obj = construct nil, false
 
               extend_object obj
 
               obj
-            when ?C
+            when 67   # ?C
               name = get_symbol
               @user_class = name
 
               construct nil, false
 
-            when ?I
+            when 73   # ?I
               ivar_index = @has_ivar.length
               @has_ivar.push true
 
@@ -394,7 +562,25 @@ module Marshal
 
       call obj if @proc and call_proc
 
-      @stream.tainted? ? obj.taint : obj
+      @stream.tainted? && !obj.frozen? ? obj.taint : obj
+    end
+
+    def construct_class
+      obj = const_lookup(get_byte_sequence.to_sym, Class)
+      store_unique_object obj
+      obj
+    end
+
+    def construct_module
+      obj = const_lookup(get_byte_sequence.to_sym, Module)
+      store_unique_object obj
+      obj
+    end
+
+    def construct_old_module
+      obj = const_lookup(get_byte_sequence.to_sym)
+      store_unique_object obj
+      obj
     end
 
     def construct_array
@@ -411,22 +597,22 @@ module Marshal
         end
       end
 
-      construct_integer.times do
-        obj << construct
+      construct_integer.times do |i|
+        obj.__append__ construct
       end
 
       obj
     end
 
     def construct_bignum
-      sign = consume_byte() == ?- ? -1 : 1
+      sign = consume_byte() == 45 ? -1 : 1  # ?-
       size = construct_integer * 2
 
       result = 0
 
       data = consume size
       (0...size).each do |exp|
-        result += (data[exp] * 2**(exp*8))
+        result += (data.getbyte(exp) * 2**(exp*8))
       end
 
       obj = result * sign
@@ -434,6 +620,27 @@ module Marshal
       store_unique_object obj
     end
 
+    def construct_data
+      name = get_symbol
+      klass = const_lookup name, Class
+      store_unique_object klass
+
+      obj = klass.allocate
+
+      # TODO ensure obj is a wrapped C pointer (T_DATA in MRI-land)
+
+      store_unique_object obj
+
+      unless Rubinius::Type.object_respond_to? obj, :_load_data
+        raise TypeError,
+              "class #{name} needs to have instance method `_load_data'"
+      end
+
+      obj._load_data construct
+
+      obj
+    end
+
     def construct_float
       s = get_byte_sequence
 
@@ -493,7 +700,7 @@ module Marshal
       return 0 if c == 0
       return c - 5 if 4 < c and c < 128
 
-      # negative, but checked known it's instead in 2's compliment
+      # negative, but checked known it's instead in 2's complement
       return c - 251 if 252 > c and c > 127
 
       # otherwise c (now in the 1 to 4 range) indicates how many
@@ -533,13 +740,17 @@ module Marshal
 
     def construct_object
       name = get_symbol
-      klass = const_lookup name
+      klass = const_lookup name, Class
       obj = klass.allocate
 
       raise TypeError, 'dump format error' unless Object === obj
 
       store_unique_object obj
-      set_instance_variables obj
+      if Rubinius::Type.object_kind_of? obj, Exception
+        set_exception_variables obj
+      else
+        set_instance_variables obj
+      end
 
       obj
     end
@@ -555,13 +766,6 @@ module Marshal
       store_unique_object obj
     end
 
-    def construct_string
-      obj = get_byte_sequence
-      obj = get_user_class.new obj if @user_class
-
-      store_unique_object obj
-    end
-
     def construct_struct
       symbols = []
       values = []
@@ -569,7 +773,7 @@ module Marshal
       name = get_symbol
       store_unique_object name
 
-      klass = const_lookup name
+      klass = const_lookup name, Class
       members = klass.members
 
       obj = klass.allocate
@@ -577,7 +781,7 @@ module Marshal
 
       construct_integer.times do |i|
         slot = get_symbol
-        unless members[i].intern == slot then
+        unless members[i].intern == slot
           raise TypeError, "struct %s is not compatible (%p for %p)" %
             [klass, slot, members[i]]
         end
@@ -597,16 +801,23 @@ module Marshal
 
     def construct_user_defined(ivar_index)
       name = get_symbol
-      klass = const_lookup name
+      klass = const_lookup name, Class
 
       data = get_byte_sequence
 
-      if ivar_index and @has_ivar[ivar_index] then
+      if Rubinius::Type.object_respond_to? klass, :__construct__
+        return klass.__construct__(self, data, ivar_index, @has_ivar)
+      end
+
+      if ivar_index and @has_ivar[ivar_index]
         set_instance_variables data
         @has_ivar[ivar_index] = false
       end
 
-      obj = klass._load data
+      obj = nil
+      Rubinius.privately do
+        obj = klass._load data
+      end
 
       store_unique_object obj
 
@@ -617,37 +828,25 @@ module Marshal
       name = get_symbol
       store_unique_object name
 
-      klass = const_lookup name
+      klass = const_lookup name, Class
       obj = klass.allocate
 
       extend_object obj if @modules
 
-      unless obj.respond_to? :marshal_load then
+      unless Rubinius::Type.object_respond_to_marshal_load? obj
         raise TypeError, "instance of #{klass} needs to have method `marshal_load'"
       end
 
       store_unique_object obj
 
       data = construct
-      obj.marshal_load data
+      Rubinius.privately do
+        obj.marshal_load data
+      end
 
       obj
     end
 
-    def consume(bytes)
-      raise ArgumentError, "marshal data too short" if @consumed > @stream.size
-      data = @stream[@consumed, bytes]
-      @consumed += bytes
-      data
-    end
-
-    def consume_byte
-      raise ArgumentError, "marshal data too short" if @consumed > @byte_array.size
-      data = @byte_array[@consumed]
-      @consumed += 1
-      return data
-    end
-
     def extend_object(obj)
       obj.__extend__(@modules.pop) until @modules.empty?
     end
@@ -666,7 +865,7 @@ module Marshal
     end
 
     def get_user_class
-      cls = const_lookup @user_class
+      cls = const_lookup @user_class, Class
       @user_class = nil
       cls
     end
@@ -675,12 +874,12 @@ module Marshal
       type = consume_byte()
 
       case type
-      when TYPE_SYMBOL then
+      when 58 # TYPE_SYMBOL
         @call = false
         obj = construct_symbol
         @call = true
         obj
-      when TYPE_SYMLINK then
+      when 59 # TYPE_SYMLINK
         num = construct_integer
         @symbols[num]
       else
@@ -699,14 +898,14 @@ module Marshal
       @depth -= 1;
 
       if link = find_link(obj)
-        str = "@#{serialize_integer(link)}"
+        str = Rubinius::Type.binary_string("@#{serialize_integer(link)}")
       else
         add_object obj
 
         # ORDER MATTERS.
-        if obj.respond_to? :marshal_dump then
+        if Rubinius::Type.object_respond_to_marshal_dump? obj
           str = serialize_user_marshal obj
-        elsif obj.respond_to? :_dump then
+        elsif Rubinius::Type.object_respond_to__dump? obj
           str = serialize_user_defined obj
         else
           str = obj.__marshal__ self
@@ -715,7 +914,7 @@ module Marshal
 
       @depth += 1
 
-      obj.tainted? ? str.taint : str
+      Rubinius::Type.infect(str, obj)
     end
 
     def serialize_extended_object(obj)
@@ -725,74 +924,57 @@ module Marshal
           str << "e#{serialize(mod.name.to_sym)}"
         end
       end
-      str
+      Rubinius::Type.binary_string(str)
     end
 
-    def serialize_mantissa(flt)
-      str = ""
-      flt = Math.modf(Math.ldexp(Math.frexp(flt.abs)[0], 37))[0]
-      if flt > 0
-        str = "\0" * 32
-        i = 0
-        while flt > 0
-          flt, n = Math.modf(Math.ldexp(flt, 32))
-          n = n.to_i
-          str[i += 1] = (n >> 24) & 0xff
-          str[i += 1] = (n >> 16) & 0xff
-          str[i += 1] = (n >> 8) & 0xff
-          str[i += 1] = (n & 0xff)
-        end
-        str.gsub!(/(\000)*\Z/, '')
-      end
-      str
+    def serializable_instance_variables(obj, exclude_ivars)
+      ivars = obj.__instance_variables__
+      ivars -= exclude_ivars if exclude_ivars
+      ivars
     end
 
     def serialize_instance_variables_prefix(obj, exclude_ivars = false)
-      ivars = obj.__instance_variables__
-
-      ivars -= exclude_ivars if exclude_ivars
-
-      if ivars.length > 0 then
-        "I"
-      else
-        ''
-      end
+      ivars = serializable_instance_variables(obj, exclude_ivars)
+      Rubinius::Type.binary_string(!ivars.empty? || serialize_encoding?(obj) ? "I" : "")
     end
 
     def serialize_instance_variables_suffix(obj, force=false,
                                             strip_ivars=false,
                                             exclude_ivars=false)
-      ivars = obj.__instance_variables__
+      ivars = serializable_instance_variables(obj, exclude_ivars)
 
-      ivars -= exclude_ivars if exclude_ivars
-
-      if force or !ivars.empty?
-        str = serialize_integer(ivars.size)
-        ivars.each do |ivar|
-          sym = ivar.to_sym
-          val = obj.__instance_variable_get__(sym)
-          if strip_ivars
-            str << serialize(ivar[1..-1].to_sym)
-          else
-            str << serialize(sym)
-          end
-          str << serialize(val)
-        end
-        str
-      else
-      ''
+      unless force or !ivars.empty? or serialize_encoding?(obj)
+        return Rubinius::Type.binary_string("")
       end
+
+      count = ivars.size
+
+      if serialize_encoding?(obj)
+        str = serialize_integer(count + 1)
+        str << serialize_encoding(obj)
+      else
+        str = serialize_integer(count)
+      end
+
+      ivars.each do |ivar|
+        sym = ivar.to_sym
+        val = obj.__instance_variable_get__(sym)
+        if strip_ivars
+          str << serialize(ivar.to_s[1..-1].to_sym)
+        else
+          str << serialize(sym)
+        end
+        str << serialize(val)
+      end
+
+      Rubinius::Type.binary_string(str)
     end
 
     def serialize_integer(n, prefix = nil)
-      if !Rubinius::L64 && n.is_a?(Fixnum)
-        prefix.to_s + serialize_fixnum(n)
+      if (!Rubinius::L64 && n.is_a?(Fixnum)) || ((n >> 31) == 0 or (n >> 31) == -1)
+        Rubinius::Type.binary_string(prefix.to_s + serialize_fixnum(n))
       else
-        if (n >> 31) == 0 or (n >> 31) == -1
-          prefix.to_s + serialize_fixnum(n)
-        else
-          serialize_bignum(n)
-        end
+        serialize_bignum(n)
       end
     end
 
@@ -814,45 +996,62 @@ module Marshal
         end
         s[0] = (n < 0 ? 256 - cnt : cnt).chr
       end
-      s
+      Rubinius::Type.binary_string(s)
     end
 
     def serialize_bignum(n)
       str = (n < 0 ? 'l-' : 'l+')
       cnt = 0
       num = n.abs
- 
+
       while num != 0
         str << (num & 0xff).chr
         num >>= 8
         cnt += 1
       end
- 
+
       if cnt % 2 == 1
         str << "\0"
         cnt += 1
       end
- 
-      str[0..1] + serialize_fixnum(cnt / 2) + str[2..-1]
+
+      Rubinius::Type.binary_string(str[0..1] + serialize_fixnum(cnt / 2) + str[2..-1])
+    end
+
+    def serialize_symbol(obj)
+      str = obj.to_s
+      Rubinius::Type.binary_string(":#{serialize_integer(str.bytesize)}#{str}")
+    end
+
+    def serialize_string(str)
+      output = Rubinius::Type.binary_string("\"#{serialize_integer(str.bytesize)}")
+      output + Rubinius::Type.binary_string(str.dup)
     end
 
     def serialize_user_class(obj, cls)
       if obj.class != cls
-        "C#{serialize(obj.class.name.to_sym)}"
+        Rubinius::Type.binary_string("C#{serialize(obj.class.name.to_sym)}")
       else
-        ''
+        Rubinius::Type.binary_string('')
       end
     end
 
     def serialize_user_defined(obj)
-      str = obj._dump @depth
+      if Rubinius::Type.object_respond_to? obj, :__custom_marshal__
+        return obj.__custom_marshal__(self)
+      end
 
-      unless str.kind_of? String
+      str = nil
+      Rubinius.privately do
+        str = obj._dump @depth
+      end
+
+      unless Rubinius::Type.object_kind_of? str, String
         raise TypeError, "_dump() must return string"
       end
 
       out = serialize_instance_variables_prefix(str)
-      out << "u#{serialize(obj.class.name.to_sym)}"
+      out << Rubinius::Type.binary_string("u#{serialize(obj.class.name.to_sym)}")
       out << serialize_integer(str.length) + str
       out << serialize_instance_variables_suffix(str)
 
@@ -860,19 +1059,16 @@ module Marshal
     end
 
     def serialize_user_marshal(obj)
-      val = obj.marshal_dump
+      val = nil
+      Rubinius.privately do
+        val = obj.marshal_dump
+      end
 
       add_object val
 
-      "U#{serialize(obj.class.__name__.to_sym)}#{val.__marshal__(self)}"
-    end
-
-    def set_instance_variables(obj)
-      construct_integer.times do
-        ivar = get_symbol
-        value = construct
-        obj.__instance_variable_set__ prepare_ivar(ivar), value
-      end
+      cls = Rubinius::Type.object_class obj
+      name = Rubinius::Type.module_inspect cls
+      Rubinius::Type.binary_string("U#{serialize(name.to_sym)}#{val.__marshal__(self)}")
     end
 
     def store_unique_object(obj)
@@ -884,11 +1080,61 @@ module Marshal
       obj
     end
 
+    def set_exception_variables(obj)
+      construct_integer.times do
+        ivar = get_symbol
+        value = construct
+        case ivar
+        when :bt
+          obj.__instance_variable_set__ :@custom_backtrace, value
+        when :mesg
+          obj.__instance_variable_set__ :@reason_message, value
+        end
+      end
+    end
+
+  end
+
+  class IOState < State
+    def consume(bytes)
+      @stream.read(bytes)
+    end
+
+    def consume_byte
+      b = @stream.getbyte
+      raise EOFError unless b
+      b
+    end
+  end
+
+  class StringState < State
+    def initialize(stream, depth, prc)
+      super stream, depth, prc
+
+      if @stream
+        @byte_array = stream.data
+      end
+
+    end
+
+    def consume(bytes)
+      raise ArgumentError, "marshal data too short" if @consumed > @stream.bytesize
+      data = @stream.byteslice @consumed, bytes
+      @consumed += bytes
+      data
+    end
+
+    def consume_byte
+      raise ArgumentError, "marshal data too short" if @consumed >= @stream.bytesize
+      data = @byte_array.get_byte @consumed
+      @consumed += 1
+      return data
+    end
   end
 
   def self.dump(obj, an_io=nil, limit=nil)
     unless limit
-      if an_io.kind_of? Fixnum
+      if Rubinius::Type.object_kind_of? an_io, Fixnum
         limit = an_io
         an_io = nil
       else
@@ -899,11 +1145,16 @@ module Marshal
     depth = Rubinius::Type.coerce_to limit, Fixnum, :to_int
     ms = State.new nil, depth, nil
 
-    if an_io and !an_io.respond_to? :write
-      raise TypeError, "output must respond to write"
+    if an_io
+      if !Rubinius::Type.object_respond_to? an_io, :write
+        raise TypeError, "output must respond to write"
+      end
+      if Rubinius::Type.object_respond_to? an_io, :binmode
+        an_io.binmode
+      end
     end
 
-    str = VERSION_STRING + ms.serialize(obj)
+    str = Rubinius::Type.binary_string(VERSION_STRING) + ms.serialize(obj)
 
     if an_io
       an_io.write(str)
@@ -913,32 +1164,30 @@ module Marshal
     return str
   end
 
-  def self.load(obj, proc = nil)
-    if obj.respond_to? :to_str
+  def self.load(obj, prc = nil)
+    if Rubinius::Type.object_respond_to? obj, :to_str
       data = obj.to_s
-    elsif obj.respond_to? :read
-      data = obj.read
-      if data.empty?
-        raise EOFError, "end of file reached"
-      end
-    elsif obj.respond_to? :getc  # FIXME - don't read all of it upfront
-      data = ''
-      data << c while (c = obj.getc.chr)
+
+      major = data.getbyte 0
+      minor = data.getbyte 1
+
+      ms = StringState.new data, nil, prc
+
+    elsif Rubinius::Type.object_respond_to? obj, :read and
+          Rubinius::Type.object_respond_to? obj, :getc
+      ms = IOState.new obj, nil, prc
+
+      major = ms.consume_byte
+      minor = ms.consume_byte
     else
       raise TypeError, "instance of IO needed"
     end
 
-    major = data[0]
-    minor = data[1]
-
-    if major != MAJOR_VERSION or minor > MINOR_VERSION then
-      raise TypeError, "incompatible marshal file format (can't be read)\n\tformat version #{MAJOR_VERSION}.#{MINOR_VERSION} required; #{major}.#{minor} given"
+    if major != MAJOR_VERSION or minor > MINOR_VERSION
+      raise TypeError, "incompatible marshal file format (can't be read)\n\tformat version #{MAJOR_VERSION}.#{MINOR_VERSION} required; #{major.inspect}.#{minor.inspect} given"
     end
 
-    ms = State.new data, nil, proc
     ms.construct
-  rescue NameError => e
-    raise ArgumentError, e.message
   end
 
   class << self
@@ -946,4 +1195,3 @@ module Marshal
   end
 
 end
-
