@@ -1,3 +1,265 @@
+module Rubinius
+  module Spawn
+    # Turns the various varargs incantations supported by Process.spawn into a
+    # [env, prog, argv, redirects, options] tuple.
+    #
+    # The following method signature is supported:
+    #   Process.spawn([env], command, ..., [options])
+    #
+    # The env and options hashes are optional. The command may be a variable
+    # number of strings or an Array full of strings that make up the new process's
+    # argv.
+    #
+    # Returns an [env, prog, argv, redirects, options] tuple. All elements are
+    # guaranteed to be non-nil. When no env or options are given, empty hashes
+    # are returned.
+    def self.extract_arguments(*args)
+      if options = Rubinius::Type.try_convert(args.last, Hash, :to_hash)
+        args.pop
+      else
+        options = {}
+      end
+
+      if env = Rubinius::Type.try_convert(args.first, Hash, :to_hash)
+        args.shift
+      else
+        env = {}
+      end
+
+      env         = adjust_env(env)
+      prog, argv  = adjust_argv(*args)
+      redir, opts = adjust_options(options)
+
+      [env, prog, argv, redir, opts]
+    end
+
+    def self.adjust_env(env)
+      adjusted = {}
+
+      env.each do |key, value|
+        key   = Rubinius::Type.check_null_safe(StringValue(key))
+        value = Rubinius::Type.check_null_safe(StringValue(value)) unless value.nil?
+
+        if key.include?("=")
+          raise ArgumentError, "environment name contains a equal : #{key}"
+        end
+
+        adjusted[key] = value
+      end
+
+      adjusted
+    end
+
+    # Mapping of string open modes to integer oflag versions.
+    OFLAGS = {
+      "r"  => File::RDONLY,
+      "r+" => File::RDWR   | File::CREAT,
+      "w"  => File::WRONLY | File::CREAT  | File::TRUNC,
+      "w+" => File::RDWR   | File::CREAT  | File::TRUNC,
+      "a"  => File::WRONLY | File::APPEND | File::CREAT,
+      "a+" => File::RDWR   | File::APPEND | File::CREAT
+    }
+
+    def self.adjust_options(options)
+      redirects = {}
+      others = {}
+
+      options.each do |key, value|
+        case key
+          when Array
+            key.each do |fd|
+              redirects[fd] = adjust_redirect_value(fd, value)
+            end
+          when :in, :out, :err, Fixnum, IO
+            redirects[key] = adjust_redirect_value(key, value)
+          when :pgroup
+            if value == true
+              value = 0
+            elsif value
+              value = Rubinius::Type.coerce_to value, Integer, :to_int
+              raise ArgumentError, "negative process group ID : #{value}" if value < 0
+            end
+            others[key] = value
+          when :chdir
+            others[key] = Rubinius::Type.coerce_to_path(value)
+          when :unsetenv_others, :umask, :close_others
+            others[key] = value
+          when Symbol
+            raise ArgumentError, "wrong exec option symbol : #{key.inspect}"
+          else
+            raise ArgumentError, "wrong exec option"
+        end
+      end
+
+      [redirects, others]
+    end
+
+    # Convert variations of redirecting to a file to a standard tuple.
+    #
+    # :in   => '/some/file'   => ['/some/file', 'r', 0644]
+    # :out  => '/some/file'   => ['/some/file', 'w', 0644]
+    # :err  => '/some/file'   => ['/some/file', 'w', 0644]
+    # STDIN => '/some/file'   => ['/some/file', 'r', 0644]
+    #
+    # Returns the modified value.
+    def self.adjust_redirect_value(key, value)
+      case value
+        when :in, :out, :err, :close, IO, Fixnum
+          value
+        when Array
+          if value.size < 3
+            defaults = default_file_reopen_info(key, value[0])
+            value += defaults[value.size..-1]
+          end
+
+          if value[1].respond_to?(:to_str)
+            value = value.dup
+            value[1] = OFLAGS[value[1]]
+          end
+
+          value
+        when String
+          default_file_reopen_info(key, value)
+        when Symbol
+          raise ArgumentError, "wrong exec redirect symbol : #{value.inspect}"
+        else
+          raise ArgumentError, "wrong exec redirect action"
+      end
+    end
+
+    # The default [file, flags, mode] tuple for a given fd and filename. The
+    # default flags vary based on the what fd is being redirected. stdout and
+    # stderr default to write, while stdin and all other fds default to read.
+    #
+    # fd   - The file descriptor that is being redirected. This may be an IO
+    #        object, integer fd number, or :in, :out, :err for one of the standard
+    #        streams.
+    # file - The string path to the file that fd should be redirected to.
+    #
+    # Returns a [file, flags, mode] tuple.
+    def self.default_file_reopen_info(fd, file)
+      case fd
+      when :in, STDIN, $stdin, 0
+        [file, OFLAGS["r"], 0644]
+      when :out, STDOUT, $stdout, 1
+        [file, OFLAGS["w"], 0644]
+      when :err, STDERR, $stderr, 2
+        [file, OFLAGS["w"], 0644]
+      else
+        [file, OFLAGS["r"], 0644]
+      end
+    end
+
+    # Convert a fd identifier to an IO object.
+    #
+    # Returns nil or an instance of IO.
+    def self.fd_to_io(object)
+      case object
+      when STDIN, STDOUT, STDERR, $stdin, $stdout, $stderr
+        object
+      when :in, 0
+        STDIN
+      when :out, 1
+        STDOUT
+      when :err, 2
+        STDERR
+      when Fixnum
+        object >= 0 ? IO.for_fd(object) : nil
+      when IO
+        object
+      else
+        object.respond_to?(:to_io) ? object.to_io : nil
+      end
+    end
+
+    # Converts the various supported command argument variations into a
+    # standard argv suitable for use with exec. This includes detecting commands
+    # to be run through the shell.
+    #
+    # The args array may follow any of these variations:
+    #
+    # 'true'                     => ['true', ['true']]
+    # 'echo', 'hello', 'world'   => ['echo', ['echo', 'hello', 'world']]
+    # 'echo *'                   => ['/bin/sh', ['sh', '-c', 'echo *']]
+    # ['echo', 'fuuu'], 'hello'  => ['echo', ['fuuu', 'hello']]
+    #
+    # Returns [prog, argv]
+    def self.adjust_argv(first, *args)
+      if args.empty? and cmd = Rubinius::Type.try_convert(first, String, :to_str)
+        if cmd.empty?
+          raise Errno::ENOENT
+        else
+          return [cmd, []]
+        end
+      elsif cmd = Rubinius::Type.try_convert(first, Array, :to_ary)
+        raise ArgumentError, "wrong first argument" unless cmd.size == 2
+        prog = StringValue(cmd[0])
+        name = StringValue(cmd[1])
+      else
+        name = prog = StringValue(first)
+      end
+
+      argv = [name]
+      args.each do |arg|
+        argv << StringValue(arg)
+      end
+
+      [prog, argv]
+    end
+
+    def self.setup_redirects(redirects)
+      redirects.each do |key, val|
+        key = fd_to_io(key)
+
+        if val == :close
+          key.close_on_exec = true
+        elsif val.is_a?(Array)
+          file, mode_string, perms = *val
+          key.reopen(File.open(file, mode_string, perms))
+        else
+          val = fd_to_io(val)
+          val.close_on_exec = false
+          val.autoclose = false
+          key.reopen(val)
+        end
+      end
+    end
+
+    def self.setup_options(options)
+      ENV.clear if options[:unsetenv_others]
+
+      if chdir = options[:chdir]
+        Dir.chdir(chdir)
+      end
+
+      # { :pgroup => pgid } options
+      pgroup = options[:pgroup]
+      pgroup = 0 if pgroup == true
+      Process.setpgid(0, pgroup) if pgroup
+
+      if umask = options[:umask]
+        File.umask(umask)
+      end
+
+      unless options[:close_others] == false
+        3.upto(IO.max_open_fd).each do |fd|
+          begin
+            IO.for_fd(fd, :autoclose => false).close_on_exec = true
+          rescue Errno::EBADF
+          end
+        end
+      end
+    end
+
+    def self.exec(env, prog, argv, redirects, options)
+      setup_options(options)
+      setup_redirects(redirects)
+      ENV.update(env)
+      Process.perform_exec prog, argv
+    end
+  end
+end
+
 module Process
   module Constants
     EXIT_SUCCESS = Rubinius::Config['rbx.platform.process.EXIT_SUCCESS'] || 0
@@ -23,26 +285,34 @@ module Process
     RLIMIT_SBSIZE  = Rubinius::Config['rbx.platform.process.RLIMIT_SBSIZE']
     RLIMIT_STACK   = Rubinius::Config['rbx.platform.process.RLIMIT_STACK']
 
+    RLIMIT_RTPRIO     = Rubinius::Config['rbx.platform.process.RLIMIT_RTPRIO']
+    RLIMIT_RTTIME     = Rubinius::Config['rbx.platform.process.RLIMIT_RTTIME']
+    RLIMIT_SIGPENDING = Rubinius::Config['rbx.platform.process.RLIMIT_SIGPENDING']
+    RLIMIT_MSGQUEUE   = Rubinius::Config['rbx.platform.process.RLIMIT_MSGQUEUE']
+    RLIMIT_NICE       = Rubinius::Config['rbx.platform.process.RLIMIT_NICE']
+
     WNOHANG = 1
     WUNTRACED = 2
   end
   include Constants
 
+  FFI = Rubinius::FFI
+
   class Rlimit < FFI::Struct
     config "rbx.platform.rlimit", :rlim_cur, :rlim_max
   end
 
   def self.setrlimit(resource, cur_limit, max_limit=undefined)
-    resource =  Rubinius::Type.coerce_to resource, Integer, :to_int
+    resource =  coerce_rlimit_resource(resource)
     cur_limit = Rubinius::Type.coerce_to cur_limit, Integer, :to_int
 
-    unless max_limit.equal? undefined
+    unless undefined.equal? max_limit
       max_limit = Rubinius::Type.coerce_to max_limit, Integer, :to_int
     end
 
     rlimit = Rlimit.new
     rlimit[:rlim_cur] = cur_limit
-    rlimit[:rlim_max] = max_limit.equal?(undefined) ? cur_limit : max_limit
+    rlimit[:rlim_max] = undefined.equal?(max_limit) ? cur_limit : max_limit
 
     ret = FFI::Platform::POSIX.setrlimit(resource, rlimit.pointer)
     Errno.handle if ret == -1
@@ -50,7 +320,7 @@ module Process
   end
 
   def self.getrlimit(resource)
-    resource =  Rubinius::Type.coerce_to resource, Integer, :to_int
+    resource = coerce_rlimit_resource(resource)
 
     lim_max = []
     rlimit = Rlimit.new
@@ -62,7 +332,7 @@ module Process
 
   def self.setsid
     pgid = FFI::Platform::POSIX.setsid
-    Errno.handle if -1 == pgid
+    Errno.handle if pgid == -1
     pgid
   end
 
@@ -76,7 +346,7 @@ module Process
       rescue SystemExit => e
         status = e.status
       rescue Exception => e
-        e.render "An exception occured in a forked block"
+        e.render "An exception occurred in a forked block"
         status = 1
       end
 
@@ -104,45 +374,50 @@ module Process
     Struct::Tms.new(*cpu_times)
   end
 
-  def self.kill(sig, pid)
+  def self.kill(signal, *pids)
+    raise ArgumentError, "PID argument required" if pids.length == 0
+
     use_process_group = false
-    sig = sig.to_s if sig.kind_of?(Symbol)
+    signal = signal.to_s if signal.kind_of?(Symbol)
 
-    if sig.kind_of?(String)
-      if sig[0] == ?-
-        sig = sig[1..-1]
+    if signal.kind_of?(String)
+      if signal[0] == ?-
+        signal = signal[1..-1]
         use_process_group = true
       end
-      if sig[0..2] == "SIG"
-        sig = sig[3..-1]
-      end
-      number = Signal::Names[sig]
-    else
-      number = sig.to_i
-      if number < 0
-        number = -number
-        use_process_group = true
+
+      if signal[0..2] == "SIG"
+        signal = signal[3..-1]
       end
+
+      signal = Signal::Names[signal]
     end
 
-    raise ArgumentError unless number
+    raise ArgumentError unless signal.kind_of? Fixnum
 
-    pid = Rubinius::Type.coerce_to pid, Integer, :to_int
-
-    pid = -pid if use_process_group
-    ret = FFI::Platform::POSIX.kill(pid, number)
-
-    case ret
-    when 0
-      return 1
-    when -1
-      Errno.handle
+    if signal < 0
+      signal = -signal
+      use_process_group = true
     end
+
+    pids.each do |pid|
+      pid = Rubinius::Type.coerce_to_pid pid
+
+      pid = -pid if use_process_group
+      result = FFI::Platform::POSIX.kill(pid, signal)
+
+      Errno.handle if result == -1
+    end
+
+    return pids.length
   end
 
   def self.abort(msg=nil)
-    $stderr.puts(msg) if msg
-    raise SystemExit.new(1)
+    if msg
+      msg = StringValue(msg)
+      $stderr.puts(msg)
+    end
+    raise SystemExit.new(1, msg)
   end
 
   def self.getpgid(pid)
@@ -192,8 +467,31 @@ module Process
   end
 
   def self.uid=(uid)
+    # the 4 rescue clauses below are needed
+    # until respond_to? can be used to query the implementation of methods attached via FFI
+    # atm respond_to returns true if a method is attached but not implemented on the platform
     uid = Rubinius::Type.coerce_to uid, Integer, :to_int
-    Process::Sys.setuid uid
+    begin
+      ret = FFI::Platform::POSIX.setresuid(uid, -1, -1)
+    rescue NotImplementedError
+      begin
+        ret = FFI::Platform::POSIX.setreuid(uid, -1)
+      rescue NotImplementedError
+        begin
+          ret = FFI::Platform::POSIX.setruid(uid)
+        rescue NotImplementedError
+          if Process.euid == uid
+            ret = FFI::Platform::POSIX.setuid(uid)
+          else
+            raise NotImplementedError
+          end
+        end
+      end
+    end
+
+    Errno.handle if ret == -1
+
+    uid
   end
 
   def self.gid=(gid)
@@ -202,8 +500,31 @@ module Process
   end
 
   def self.euid=(uid)
+    # the 4 rescue clauses below are needed
+    # until respond_to? can be used to query the implementation of methods attached via FFI
+    # atm respond_to returns true if a method is attached but not implemented on the platform
     uid = Rubinius::Type.coerce_to uid, Integer, :to_int
-    Process::Sys.seteuid uid
+    begin
+      ret = FFI::Platform::POSIX.setresuid(-1, uid, -1)
+    rescue NotImplementedError
+      begin
+        ret = FFI::Platform::POSIX.setreuid(-1, uid)
+      rescue NotImplementedError
+        begin
+          ret = FFI::Platform::POSIX.seteuid(uid)
+        rescue NotImplementedError
+          if Process.uid == uid
+            ret = FFI::Platform::POSIX.setuid(uid)
+          else
+            raise NotImplementedError
+          end
+        end
+      end
+    end
+
+    Errno.handle if ret == -1
+
+    uid
   end
 
   def self.egid=(gid)
@@ -269,7 +590,7 @@ module Process
     @maxgroups = g.length if g.length > @maxgroups
     FFI::MemoryPointer.new(:int, @maxgroups) { |p|
       p.write_array_of_int(g)
-      Errno.handle if -1 == FFI::Platform::POSIX.setgroups(g.length, p)
+      Errno.handle if FFI::Platform::POSIX.setgroups(g.length, p) == -1
     }
     g
   end
@@ -332,7 +653,7 @@ module Process
     status, termsig, stopsig, pid = value
 
     status = Process::Status.new(pid, status, termsig, stopsig)
-    Rubinius::Globals.set! :$?, status
+    set_status_global status
     [pid, status]
   end
 
@@ -371,13 +692,68 @@ module Process
     alias_method :waitpid2, :wait2
   end
 
-  #
-  # Indicate disinterest in child process.
-  #
-  # Sets up an internal wait on the given process ID.
-  # Only possibly real pids, i.e. positive numbers,
-  # may be waited for.
-  #
+  Rubinius::Globals.read_only :$?
+  Rubinius::Globals.set_hook(:$?) { Thread.current[:$?] }
+
+  def self.set_status_global(status)
+    Thread.current[:$?] = status
+  end
+
+  def self.daemon(stay_in_dir=false, keep_stdio_open=false)
+    # Do not run at_exit handlers in the parent
+    exit!(0) if fork
+
+    Process.setsid
+
+    exit!(0) if fork
+
+    Dir.chdir("/") unless stay_in_dir
+
+    unless keep_stdio_open
+      io = File.open "/dev/null", File::RDWR, 0
+      $stdin.reopen io
+      $stdout.reopen io
+      $stderr.reopen io
+    end
+
+    return 0
+  end
+
+  def self.exec(*args)
+    env, prog, argv, redirects, options = Rubinius::Spawn.extract_arguments(*args)
+    Rubinius::Spawn.exec(env, prog, argv, redirects, options)
+  end
+
+  def self.spawn(*args)
+    env, prog, argv, redirects, options = Rubinius::Spawn.extract_arguments(*args)
+
+    IO.pipe do |read, write|
+      pid = Process.fork do
+        write.close_on_exec = true
+
+        begin
+          Rubinius::Spawn.exec(env, prog, argv, redirects, options)
+        rescue => e
+          write.write Marshal.dump(e)
+          exit! 1
+        end
+      end
+
+      write.close
+
+      unless read.eof?
+        error = Marshal.load(read)
+
+        # Intercepting the "command not found" error to set $?
+        set_status_global Process::Status.new(pid, 127) if Errno::ENOENT === error
+
+        raise error
+      end
+
+      pid
+    end
+  end
+
   # TODO: Should an error be raised on ECHILD? --rue
   #
   # TODO: This operates on the assumption that waiting on
@@ -388,10 +764,35 @@ module Process
   def self.detach(pid)
     raise ArgumentError, "Only positive pids may be detached" unless pid > 0
 
-    # The evented system does not need a loop
-    Thread.new { Process.wait pid }
+    thread = Thread.new { Process.wait pid; $? }
+    thread[:pid] = pid
+    def thread.pid; self[:pid] end
+
+    thread
   end
 
+  def self.coerce_rlimit_resource(resource)
+    case resource
+    when Integer
+      return resource
+    when Symbol, String
+      # do nothing
+    else
+      unless r = Rubinius::Type.check_convert_type(resource, String, :to_str)
+        return Rubinius::Type.coerce_to resource, Integer, :to_int
+      end
+
+      resource = r
+    end
+
+    constant = "RLIMIT_#{resource}"
+    unless const_defined? constant
+      raise ArgumentError, "invalid resource name: #{constant}"
+    end
+    const_get constant
+  end
+  private_class_method :coerce_rlimit_resource
+
   #--
   # TODO: Most of the fields aren't implemented yet.
   # TODO: Also, these objects should only need to be constructed by
@@ -410,48 +811,48 @@ module Process
       @termsig = termsig
       @stopsig = stopsig
     end
-    
+
     def to_i
       @exitstatus
     end
-    
+
     def to_s
       @exitstatus.to_s
     end
-    
+
     def &(num)
       @exitstatus & num
     end
-    
+
     def ==(other)
       other = other.to_i if other.kind_of? Process::Status
       @exitstatus == other
     end
-    
+
     def >>(num)
       @exitstatus >> num
     end
-    
+
     def coredump?
       false
     end
-    
+
     def exited?
       @exitstatus != nil
     end
-    
+
     def pid
       @pid
     end
-    
+
     def signaled?
       @termsig != nil
     end
-    
+
     def stopped?
       @stopsig != nil
     end
-    
+
     def success?
       if exited?
         @exitstatus == 0
@@ -540,10 +941,11 @@ module Process
         nil
       end
 
-      def setreuid(rid)
+      def setreuid(rid, eid)
         rid = Rubinius::Type.coerce_to rid, Integer, :to_int
+        eid = Rubinius::Type.coerce_to eid, Integer, :to_int
 
-        ret = FFI::Platform::POSIX.setreuid rid
+        ret = FFI::Platform::POSIX.setreuid rid, eid
         Errno.handle if ret == -1
         nil
       end
@@ -558,7 +960,7 @@ module Process
         nil
       end
 
-      def setresuid(rid, eig, sid)
+      def setresuid(rid, eid, sid)
         rid = Rubinius::Type.coerce_to rid, Integer, :to_int
         eid = Rubinius::Type.coerce_to eid, Integer, :to_int
         sid = Rubinius::Type.coerce_to sid, Integer, :to_int
@@ -699,67 +1101,34 @@ module Process
 end
 
 module Kernel
+  def system(*args)
+    begin
+      pid = Process.spawn(*args)
+    rescue SystemCallError
+      nil
+    else
+      Process.waitpid(pid)
+      $?.exitstatus == 0
+    end
+  end
+  module_function :system
+
+  def spawn(*args)
+    Process.spawn(*args)
+  end
+  module_function :spawn
 
   def fork(&block)
     Process.fork(&block)
   end
   module_function :fork
 
-  def system(prog, *args)
-    pid = Process.fork
-    if pid
-      Process.waitpid(pid)
-      $?.exitstatus == 0
-    else
-      begin
-        Kernel.exec(prog, *args)
-      rescue Exception => e
-        if $DEBUG
-          e.render("Unable to execute subprogram", STDERR)
-        end
-        exit! 1
-      end
-
-      if $DEBUG
-        STDERR.puts "Unable to execute subprogram - exec silently returned"
-      end
-      exit! 1
-    end
-  end
-  module_function :system
-
-  def exec(cmd, *args)
-    if args.empty? and cmd.kind_of? String
-      raise Errno::ENOENT if cmd.empty?
-      if /([*?{}\[\]<>()~&|$;'`"\n\s]|[^\w])/o.match(cmd)
-        Process.perform_exec "/bin/sh", ["sh", "-c", cmd]
-      else
-        Process.perform_exec cmd, [cmd]
-      end
-    else
-      if cmd.kind_of? Array
-        prog = cmd[0]
-        name = cmd[1]
-      else
-        name = prog = cmd
-      end
-
-      argv = [name]
-      args.each do |arg|
-        argv << arg.to_s
-      end
-
-      Process.perform_exec prog, argv
-    end
-  end
-  module_function :exec
-
   def `(str) #`
     str = StringValue(str) unless str.kind_of?(String)
     pid, output = Process.replace(str)
 
     Process.waitpid(pid)
-    return output
+    return Rubinius::Type.external_string(output)
   end
 
   module_function :` # `
