@@ -1,26 +1,31 @@
 TOPLEVEL_BINDING = binding()
 
-# Default kcode
-Rubinius.kcode = "ASCII"
-
 module Rubinius
   class Loader
     def initialize
       @exit_code    = 0
+      @process_id   = Process.pid
       @load_paths   = []
       @requires     = []
       @evals        = []
       @script       = nil
       @debugging    = false
-      @run_irb      = true
+      @repl         = true
       @printed_version = false
       @input_loop   = false
       @input_loop_print = false
       @input_loop_split = false
       @simple_options = false
       @early_option_stop = false
+      @check_syntax = false
+      @load_profiler = false
+      @enable_gems = true
+      @load_gemfile = false
 
-      @gem_bin = File.join Rubinius::GEMS_PATH, "bin"
+      version = RUBY_VERSION.split(".").first(2).join(".")
+      @gem_bins = ["#{version}/bin", "bin"].map do |dir|
+        File.join Rubinius::GEMS_PATH, dir
+      end
     end
 
     def self.debugger
@@ -33,22 +38,13 @@ module Rubinius
 
     # Finish setting up after loading kernel.
     def preamble
+      set_program_name "rbx"
       @stage = "running Loader preamble"
 
       Object.const_set :ENV, EnvironmentVariables.new
 
-      # set terminal width
-      width = 80
-      if Terminal and ENV['TERM'] and !ENV['RBX_NO_COLS']
-        begin
-          `which tput &> /dev/null`
-          if $?.exitstatus == 0
-            res = `tput cols 2>&1`.to_i
-            width = res if res > 0
-          end
-        end
-      end
-      Rubinius.const_set 'TERMINAL_WIDTH', width
+      # Set the default visibility for the top level binding
+      TOPLEVEL_BINDING.variables.method_visibility = :private
 
       $VERBOSE = false
     end
@@ -57,44 +53,24 @@ module Rubinius
     def system_load_path
       @stage = "setting up system load path"
 
-      @main_lib = nil
-
-      if env_lib = ENV['RBX_LIB']
-        @main_lib = env_lib if File.exists?(env_lib)
-      end
-
-      # Use the env version if it's set.
-      @main_lib = Rubinius::LIB_PATH unless @main_lib
-
-      unless @main_lib
-        STDERR.puts <<-EOM
-Rubinius was configured to find standard library files at:
-
-  #{@main_lib}
-
-but that directory does not exist.
-
-Set the environment variable RBX_LIB to the directory
-containing the Rubinius standard library files.
-        EOM
-      end
-
+      @main_lib = Rubinius::LIB_PATH
       @main_lib_bin = File.join @main_lib, "bin"
 
       # This conforms more closely to MRI. It is necessary to support
       # paths that mkmf adds when compiling and installing native exts.
-      additions = []
-      additions << Rubinius::SITE_PATH
-      additions << "#{Rubinius::SITE_PATH}/#{Rubinius::CPU}-#{Rubinius::OS}"
-      additions << Rubinius::VENDOR_PATH
-      additions << "#{Rubinius::VENDOR_PATH}/#{Rubinius::CPU}-#{Rubinius::OS}"
-      additions << @main_lib
+      additions = [
+        Rubinius::SITE_PATH,
+        "#{Rubinius::SITE_PATH}/#{RUBY_VERSION}",
+        "#{Rubinius::SITE_PATH}/#{Rubinius::CPU}-#{Rubinius::OS}",
+        Rubinius::VENDOR_PATH,
+        @main_lib,
+      ]
       additions.uniq!
 
       $LOAD_PATH.unshift(*additions)
 
       if ENV['RUBYLIB'] and not ENV['RUBYLIB'].empty? then
-        rubylib_paths = ENV['RUBYLIB'].split(':')
+        rubylib_paths = ENV['RUBYLIB'].split(File::PATH_SEPARATOR)
         $LOAD_PATH.unshift(*rubylib_paths)
       end
     end
@@ -123,12 +99,17 @@ containing the Rubinius standard library files.
       Signal.trap("INT") do |sig|
         raise Interrupt, "Thread has been interrupted"
       end
+
+      ["HUP", "QUIT", "TERM", "ALRM", "USR1", "USR2"].each do |signal|
+        Signal.trap(signal) do |sig|
+          raise SignalException, sig
+        end
+      end
     end
 
     def show_syntax_error(e)
       STDERR.puts "A syntax error has occurred:"
-      STDERR.puts "    #{e.reason}"
-      STDERR.puts "    near line #{e.file}:#{e.line}, column #{e.column}"
+      STDERR.puts "    #{e.message}"
       STDERR.puts "\nCode:\n#{e.code}"
       if e.column
         STDERR.puts((" " * (e.column - 1)) + "^")
@@ -141,7 +122,7 @@ containing the Rubinius standard library files.
       end
     end
 
-    # Checks if a subcammand with basename +base+ exists. Returns the full
+    # Checks if a subcommand with basename +base+ exists. Returns the full
     # path to the subcommand if it does; otherwise, returns nil.
     def find_subcommand(base)
       command = File.join @main_lib_bin, "#{base}.rb"
@@ -152,8 +133,10 @@ containing the Rubinius standard library files.
     # Checks if a gem wrapper named +base+ exists. Returns the full path to
     # the gem wrapper if it does; otherwise, returns nil.
     def find_gem_wrapper(base)
-      wrapper = File.join @gem_bin, base
-      return wrapper if File.exists? wrapper
+      @gem_bins.each do |dir|
+        wrapper = File.join dir, base
+        return wrapper if File.exists? wrapper
+      end
       return nil
     end
 
@@ -202,6 +185,36 @@ containing the Rubinius standard library files.
       end
     end
 
+    def define_global_methods
+      Kernel.module_eval do
+        def gsub(pattern, rep=nil, &block)
+          target = $_
+          raise TypeError, "$_ must be a String, but is #{target.inspect}" unless target.kind_of? String
+          $_ = target.gsub(pattern, rep, &block)
+        end
+        module_function :gsub
+
+        def sub(pattern, rep=nil, &block)
+          target = $_
+          raise TypeError, "$_ must be a String, but is #{target.inspect}" unless target.kind_of? String
+          $_ = target.sub(pattern, rep, &block)
+        end
+        module_function :sub
+
+        def chomp(string=$/)
+          raise TypeError, "$_ must be a String" unless $_.kind_of? String
+          $_ = $_.chomp(string)
+        end
+        module_function :chomp
+
+        def chop
+          raise TypeError, "$_ must be a String" unless $_.kind_of? String
+          $_ = $_.chop
+        end
+        module_function :chop
+      end
+    end
+
     # Process all command line arguments.
     def options(argv=ARGV)
       @stage = "processing command line arguments"
@@ -223,9 +236,12 @@ containing the Rubinius standard library files.
 
       options.doc "\nRuby options"
       options.on "-", "Read and evaluate code from STDIN" do
-        @run_irb = false
-        $0 = "-"
-        CodeLoader.execute_script STDIN.read
+        @repl = false
+        set_program_name "-"
+        stdin = STDIN.dup
+        script = STDIN.read
+        STDIN.reopen(stdin)
+        CodeLoader.execute_script script
       end
 
       options.on "--", "Stop processing command line arguments" do
@@ -235,25 +251,12 @@ containing the Rubinius standard library files.
 
       options.on "-a", "Used with -n and -p, splits $_ into $F" do
         @input_loop_split = true
+        Rubinius::Globals.set!(:$-a, true)
       end
 
-      options.on "-c", "FILE", "Check the syntax of FILE" do |file|
-        if File.exists?(file)
-          mel = Rubinius::Melbourne.new file, 1, []
-
-          begin
-            mel.parse_file
-          rescue SyntaxError => e
-            show_syntax_errors(mel.syntax_errors)
-            exit 1
-          end
-
-          puts "Syntax OK"
-          exit 0
-        else
-          puts "rbx: Unable to find file -- #{file} (LoadError)"
-          exit 1
-        end
+      options.on "-c", "Only check the syntax" do
+        @repl = false
+        @check_syntax = true
       end
 
       options.on "-C", "DIR", "Change directory to DIR before running scripts" do |dir|
@@ -261,17 +264,30 @@ containing the Rubinius standard library files.
       end
 
       options.on "-d", "Enable debugging output and set $DEBUG to true" do
+        $VERBOSE = true
         $DEBUG = true
       end
 
+      options.on "--disable-gems", "Do not automatically load rubygems on startup" do
+        @enable_gems = false
+      end
+
       options.on "-e", "CODE", "Compile and execute CODE" do |code|
-        @run_irb = false
-        $0 = "(eval)"
+        @repl = false
+        set_program_name "(eval)"
         @evals << code
       end
 
+      options.on '-F', "PATTERN", "Set $; to PATTERN" do |pattern|
+        Rubinius::Globals.set!(:$;, Regexp.new(pattern))
+      end
+
+      options.on '-G', '--gemfile', 'Respect a Gemfile in the current path' do
+        @load_gemfile = true
+      end
+
       options.on "-h", "--help", "Display this help" do
-        @run_irb = false
+        @repl = false
         puts options
         done
       end
@@ -286,28 +302,27 @@ containing the Rubinius standard library files.
         @load_paths << dir
       end
 
-      options.on "-K", "[code]", "Set $KCODE" do |k|
-        case k
-        when 'a', 'A', 'n', 'N', nil
-          $KCODE = "NONE"
-        when 'e', 'E'
-          $KCODE = "EUC"
-        when 's', 'S'
-          $KCODE = "SJIS"
-        when 'u', 'U'
-          $KCODE = "UTF8"
-        else
-          $KCODE = "NONE"
-        end
+      options.on "-K", "Ignored $KCODE option for compatibility"
+      options.on "-U", "Set Encoding.default_internal to UTF-8" do
+        set_default_internal_encoding('UTF-8')
+      end
+
+      options.on "-E", "ENC", "Set external:internal character encoding to ENC" do |enc|
+        ext, int = enc.split(":")
+        Encoding.default_external = ext if ext and !ext.empty?
+        set_default_internal_encoding(int) if int and !int.empty?
       end
 
       options.on "-n", "Wrap running code in 'while(gets()) ...'" do
         @input_loop = true
+        define_global_methods
       end
 
       options.on "-p", "Same as -n, but also print $_" do
         @input_loop = true
         @input_loop_print = true
+        Rubinius::Globals.set!(:$-p, true)
+        define_global_methods
       end
 
       options.on "-r", "LIBRARY", "Require library before execution" do |file|
@@ -321,7 +336,11 @@ containing the Rubinius standard library files.
       options.on("-S", "SCRIPT",
                  "Run SCRIPT using PATH environment variable to find it") do |script|
         options.stop_parsing
-        @run_irb = false
+        @repl = false
+
+        # Load a gemfile now if we need to so that -S can see binstubs
+        # internal to the Gemfile
+        gemfile
 
         # First, check if any existing gem wrappers match.
         unless file = find_gem_wrapper(script)
@@ -339,17 +358,14 @@ containing the Rubinius standard library files.
           end
         end
 
-        $0 = script if file
+        set_program_name script if file
 
         # if missing, let it die a natural death
         @script = file ? file : script
       end
 
-      options.on "-T", "[level]", "Set $SAFE level (NOT IMPLEMENTED)" do |l|
-      end
-
       options.on "-v", "Display the version and set $VERBOSE to true" do
-        @run_irb = false
+        @repl = false
         $VERBOSE = true
 
         unless @printed_version
@@ -379,43 +395,10 @@ containing the Rubinius standard library files.
       end
 
       options.on "--version", "Display the version" do
-        @run_irb = false
+        @repl = false
         puts Rubinius.version
       end
 
-
-      # TODO: convert all these to -X options
-      options.doc "\nRubinius options"
-
-      options.on "--gc-stats", "Show GC stats" do
-        stats = Stats::GC.new
-        at_exit { stats.show }
-      end
-
-      @profile = Rubinius::Config['profile'] || Rubinius::Config['jit.profile']
-
-      options.on "--vv", "Display version and extra info" do
-        @run_irb = false
-
-        $VERBOSE = true
-        puts Rubinius.version
-        puts "Options:"
-        puts "  Interpreter type: #{INTERPRETER}"
-        if jit = JIT
-          puts "  JIT enabled: #{jit.join(', ')}"
-        else
-          puts "  JIT disabled"
-        end
-        puts
-      end
-
-      # This will only trigger if it's not the first option, in which case
-      # we'll just tell the user to make it the first option.
-      options.on "--rebuild-compiler", "Rebuild the Rubinius compiler" do
-        puts "This must be the first and only option."
-        exit 1
-      end
-
       options.doc <<-DOC
 \nRubinius subcommands
 
@@ -439,7 +422,7 @@ VM Options
      This option is recognized by the VM before any ruby code is loaded.
      It is used to set VM configuration options.
 
-     Use -Xconfig.print to see the list of options the VM recognizes.
+     Use -Xhelp to see the list of options the VM recognizes.
      All variables, even ones that the VM doesn't understand, are available
      in Rubinius::Config.
 
@@ -454,6 +437,18 @@ VM Options
         handle_simple_options(argv)
       end
 
+      if Rubinius::Config['help']
+        STDOUT.puts "Rubinius configuration variables:"
+        STDOUT.puts "  These variables are normally set via -X and read via Rubinius::Config[var]."
+        STDOUT.puts
+
+        require 'rubinius/configuration'
+        Rubinius::ConfigurationVariables.instance.show_help(STDOUT)
+        exit 0
+      end
+
+      exit 0 if Rubinius::Config["config.print"]
+
       if str = Rubinius::Config['tool.require']
         begin
           require str
@@ -462,11 +457,31 @@ VM Options
         end
       end
 
-      if @profile
-        require 'profile'
+      if Rubinius::Config['profile'] || Rubinius::Config['jit.profile']
+        @load_profiler = true
+      end
+
+      if @check_syntax
+        check_syntax
       end
     end
 
+    # Sets $0 ($PROGRAM_NAME) without changing the process title
+    def set_program_name(name)
+      Rubinius::Globals.set! :$0, name
+    end
+    private :set_program_name
+
+    def set_default_internal_encoding(encoding)
+      if @default_internal_encoding_set && Encoding.default_internal.name != encoding
+        raise RuntimeError, "Default internal encoding already set to '#{Encoding.default_internal.name}'."
+      else
+        @default_internal_encoding_set = true
+        Encoding.default_internal = encoding
+      end
+    end
+    private :set_default_internal_encoding
+
     RUBYOPT_VALID_OPTIONS = "IdvwWrKT"
 
     def handle_rubyopt(options)
@@ -506,37 +521,36 @@ VM Options
       end
     end
 
-    def load_compiler
-      @stage = "loading the compiler"
-
-      # This happens before we parse ARGV, so we have to check ARGV ourselves
-      # here.
-
-      rebuild = (ARGV.last == "--rebuild-compiler")
+    def run_compiled
+      return unless ENV["RBX_RUN_COMPILED"]
 
       begin
-        CodeLoader.require_compiled "compiler", rebuild ? false : true
-      rescue Rubinius::InvalidRBC => e
-        STDERR.puts "There was an error loading the compiler."
-        STDERR.puts "It appears that your compiler is out of date with the VM."
-        STDERR.puts "\nPlease use 'rbx --rebuild-compiler' or 'rake [install]' to"
-        STDERR.puts "bring the compiler back to date."
+        ARGV.each do |script|
+          CodeLoader.require_compiled script
+        end
+      rescue Object => e
+        STDERR.puts "Unable to run compiled file: #{script}"
+        e.render
         exit 1
       end
 
-      if rebuild
-        STDOUT.puts "Rebuilding compiler..."
-        files =
-          ["#{@main_lib}/compiler.rb"] +
-          Dir["#{@main_lib}/compiler/*.rb"] +
-          Dir["#{@main_lib}/compiler/**/*.rb"]
+      exit 0
+    end
 
-        files.each do |file|
-          puts "#{file}"
-          Rubinius.compile_file file, true
-        end
+    def load_compiler
+      @stage = "loading the compiler"
 
-        exit 0
+      begin
+        CodeLoader.load_compiler
+      rescue LoadError => e
+        STDERR.puts <<-EOM
+Unable to load the bytecode compiler. Please run 'rake' or 'rake install'
+to rebuild the compiler.
+
+        EOM
+
+        e.render
+        exit 1
       end
     end
 
@@ -553,17 +567,29 @@ VM Options
       end
     end
 
-    def agent
-      @stage = "starting agent ruby thread"
+    def rubygems
+      @stage = "loading Rubygems"
 
-      # Fixing a bug on OSX 10.5
-      return
+      CodeLoader.load_rubygems if @enable_gems
+    end
 
-      if Rubinius::Config['agent.start']
-        Rubinius::AgentRegistry.spawn_thread
+    def gemfile
+      @stage = "loading Gemfile"
+
+      if @load_gemfile
+        CodeLoader.load_rubygems
+        require 'bundler/setup'
+        @load_gemfile = false
       end
     end
 
+    def profiler
+      return unless @load_profiler
+      @stage = "loading profiler"
+
+      require 'profile'
+    end
+
     # Require any -r arguments
     def requires
       @stage = "requiring command line files"
@@ -575,13 +601,16 @@ VM Options
     def evals
       return if @evals.empty?
 
+      @repl = false
       @stage = "evaluating command line code"
 
+      Dir.chdir(@directory) if @directory
+
       if @input_loop
         while gets
           $F = $_.split if @input_loop_split
           eval @evals.join("\n"), TOPLEVEL_BINDING, "-e", 1
-          puts $_ if @input_loop_print
+          print $_ if @input_loop_print
         end
       else
         eval @evals.join("\n"), TOPLEVEL_BINDING, "-e", 1
@@ -592,6 +621,8 @@ VM Options
     def script
       return unless @script and @evals.empty?
 
+      @repl = false
+
       handle_simple_options(ARGV) if @simple_options
 
       @stage = "running #{@script}"
@@ -614,40 +645,81 @@ VM Options
         end
       end
 
-      $0 = @script
+      set_program_name @script
       CodeLoader.load_script @script, @debugging
     end
 
-    # Run IRB unless we were passed -e, -S arguments or a script to run.
-    def irb
-      return if $0 or not @run_irb
+    #Check Ruby syntax of source
+    def check_syntax
+      parser = Rubinius::ToolSet::Runtime::Melbourne
 
-      @stage = "running IRB"
+      if @script
+        if File.exists?(@script)
+          mel = parser.new @script, 1, []
 
-      if Terminal
-        repr = ENV['RBX_REPR'] || "bin/irb"
-        $0 = repr
-        prog = File.join @main_lib, repr
+          begin
+            mel.parse_file
+          rescue SyntaxError => e
+            show_syntax_errors(mel.syntax_errors)
+            exit 1
+          end
+        else
+          puts "rbx: Unable to find file -- #{@script} (LoadError)"
+          exit 1
+        end
+      elsif not @evals.empty?
         begin
-          # HACK: this was load but load raises LoadError
-          # with prog == "lib/bin/irb". However, require works.
-          # Investigate when we have specs running.
-          require prog
-        rescue LoadError => e
-          STDERR.puts "Unable to load REPL named '#{repr}'"
-          STDERR.puts e.message
-          puts e.awesome_backtrace.show
+          mel = parser.parse_string @evals.join("\n")
+        rescue SyntaxError => e
+          show_syntax_errors(mel.syntax_errors)
           exit 1
         end
       else
-        $0 = "(eval)"
+        begin
+          mel = parser.parse_string STDIN.read
+        rescue SyntaxError => e
+          show_syntax_errors(mel.syntax_errors)
+          exit 1
+        end
+      end
+
+      puts "Syntax OK"
+      exit 0
+    end
+
+    # Run IRB unless we were passed -e, -S arguments or a script to run.
+    def repl
+      return unless @repl
+
+      @stage = "running IRB"
+
+      # Check if a different REPL is set.
+      repl = ENV["RBX_REPL"] || "irb"
+
+      if Terminal
+        load File.join(Rubinius::GEMS_PATH, "bin", repl)
+      else
+        set_program_name "(eval)"
         CodeLoader.execute_script "p #{STDIN.read}"
       end
     end
 
+    def run_at_exits
+      until AtExit.empty?
+        begin
+          AtExit.shift.call
+        rescue SystemExit => e
+          @exit_code = e.status
+          # We recurse down so that future at_exits
+          # see the current exception
+          run_at_exits
+        end
+      end
+    end
+
     # Cleanup and at_exit processing.
     def epilogue
-      @stage = "at_exit handler"
+      @stage = "running at_exit handlers"
 
       begin
         Signal.run_handler Signal::Names["EXIT"]
@@ -655,16 +727,10 @@ VM Options
         @exit_code = e.status
       end
 
-      until AtExit.empty?
-        begin
-          AtExit.shift.call
-        rescue SystemExit => e
-          @exit_code = e.status
-        end
-      end
+      run_at_exits
 
-      @stage = "object finalizers"
-      GC.start
+      @stage = "running object finalizers"
+      ::GC.start
       ObjectSpace.run_finalizers
 
       # TODO: Fix these with better -X processing
@@ -672,9 +738,6 @@ VM Options
         p VM.jit_info
       end
 
-      if Config['rbx.gc_stats']
-        Stats::GC.new.show
-      end
     rescue Object => e
       e.render "An exception occurred #{@stage}"
       @exit_code = 1
@@ -689,17 +752,22 @@ VM Options
       thread_state = Rubinius.thread_state
       reason = thread_state[0]
       unless reason == :none
-        STDERR.puts "\nERROR: the VM is exiting improperly"
+        STDERR.puts "\nERROR: the VM is exiting improperly #{@stage}"
         STDERR.puts "intended operation: #{reason.inspect}"
         STDERR.puts "associated value: #{thread_state[1].inspect}"
 
         destination = thread_state[2]
-        method = destination.method
 
-        STDERR.puts "destination scope:"
-        STDERR.puts "  method: #{method.name} at #{method.file}:#{method.first_line}"
-        STDERR.puts "  module: #{destination.module.name}"
-        STDERR.puts "  block:  #{destination.block}" if destination.block
+        if destination
+          method = destination.method
+
+          STDERR.puts "destination scope:"
+          STDERR.puts "  method: #{method.name} at #{method.file}:#{method.first_line}"
+          STDERR.puts "  module: #{destination.module.name}"
+          STDERR.puts "  block:  #{destination.block}" if destination.block
+        else
+          STDERR.puts "destination scope: unknown"
+        end
 
         if reason == :catch_throw
           STDERR.puts "throw destination: #{thread_state[4].inspect}"
@@ -726,10 +794,10 @@ VM Options
 
     def write_last_error(e)
       unless path = Config['vm.crash_report_path']
-        path = "#{ENV['HOME']}/.rubinius_last_error"
+        path = "#{ENV['HOME']}/.rbx/rubinius_last_error_#{@process_id}"
       end
 
-      File.open(path, "w") do |f|
+      File.open(path, "wb") do |f|
         f.puts "Rubinius Crash Report #rbxcrashreport"
         f.puts ""
         f.puts "[[Exception]]"
@@ -745,66 +813,60 @@ VM Options
 
     # Orchestrate everything.
     def main
-      begin
-        begin
-          preamble
-          system_load_path
-          signals
-          load_compiler
-          preload
-          detect_alias
-          options
-          load_paths
-          debugger
-          agent
-          requires
-          evals
-          script
-          irb
+      preamble
+      system_load_path
+      signals
+      run_compiled
+      load_compiler
+      preload
+      detect_alias
+      options
+      load_paths
+      rubygems
+      gemfile
+      debugger
+      profiler
+      requires
+      evals
+      script
+      repl
 
-        rescue SystemExit => e
-          # Let the outer rescue grab it
-          raise e
+    rescue SystemExit => e
+      @exit_code = e.status
 
-        rescue SyntaxError => e
-          @exit_code = 1
+      epilogue
+    rescue SyntaxError => e
+      @exit_code = 1
 
-          show_syntax_error(e)
+      show_syntax_error(e)
 
-          STDERR.puts "\nBacktrace:"
-          STDERR.puts e.awesome_backtrace.show
-        rescue Object => e
-          @exit_code = 1
+      STDERR.puts "\nBacktrace:"
+      STDERR.puts
+      STDERR.puts e.awesome_backtrace.show
+      epilogue
+    rescue Interrupt => e
+      @exit_code = 1
 
-          write_last_error(e)
-          e.render "An exception occurred #{@stage}"
-        end
+      write_last_error(e)
+      e.render "An exception occurred #{@stage}:"
+      epilogue
+    rescue SignalException => e
+      Signal.trap(e.signo, "SIG_DFL")
+      Process.kill e.signo, Process.pid
+      epilogue
+    rescue Object => e
+      @exit_code = 1
 
-      # We do this, run epilogue both on catching SystemExit and
-      # if there was no exception so that the at_exit handlers
-      # can see $! as the SystemExit object the system is going to
-      # exit with.
-      rescue SystemExit => e
-        @exit_code = e.status
-        epilogue
-      else
-        epilogue
-      ensure
-        done
-      end
-    end
-
-    # Creates an instance of the Loader and runs it. We catch any uncaught
-    # exceptions here and report them before exiting.
-    def self.main
-      begin
-        new.main
-      rescue Object => exc
-        STDERR.puts "\n====================================="
-        STDERR.puts "Exception occurred during top-level exception output! (THIS IS BAD)"
-        STDERR.puts
-        STDERR.puts "Exception: #{exc.inspect} (#{exc.class})"
-      end
+      write_last_error(e)
+      e.render "An exception occurred #{@stage}:"
+      epilogue
+    else
+      # We do this, run epilogue both in the rescue blocks and also here,
+      # so that at_exit{} hooks can read $!.
+      epilogue
+    ensure
+      done
     end
   end
 end
+
